from typing import List, Optional

from firebolt.model.V2.instance_type import InstanceType
from firebolt.service.V2.base import BaseService


class InstanceTypeService(BaseService):
    @staticmethod
    def instance_types() -> List[InstanceType]:
        return [it for it in InstanceType]

    @staticmethod
    def cheapest_instance() -> Optional[InstanceType]:
        return InstanceType.S

    @staticmethod
    def get(name: str) -> Optional[InstanceType]:
        """
        Get an instance type by name.

        Args:
            name: Name of the instance (eg. "i3.4xlarge")

        Returns:
            The requested instance type or None if it wasn't found
        """
        try:
            return InstanceType(name)
        except ValueError:
            return None



from enum import Enum


class EngineStatus(Enum):
    """
    Detailed engine status.
    """

    STARTING = "STARTING"
    STARTED = "STARTED"
    RUNNING = "RUNNING"
    STOPPING = "STOPPING"
    STOPPED = "STOPPED"
    DROPPING = "DROPPING"
    REPAIRING = "REPAIRING"
    FAILED = "FAILED"
    DELETING = "DELETING"

    def __str__(self) -> str:
        return self.value



from firebolt.client.auth import Auth  # backward compatibility
from firebolt.client.client import (
    AsyncClient,
    AsyncClientV1,
    AsyncClientV2,
    Client,
    ClientV1,
    ClientV2,
)
from firebolt.client.constants import DEFAULT_API_URL
from firebolt.client.resource_manager_hooks import (
    log_request,
    log_response,
    raise_on_4xx_5xx,
)



from __future__ import annotations

import logging
import re
import time
from dataclasses import dataclass, field
from typing import TYPE_CHECKING, ClassVar, List, Optional, Tuple, Union

from firebolt.db import Connection, connect
from firebolt.model.V2 import FireboltBaseModel
from firebolt.model.V2.database import Database
from firebolt.model.V2.instance_type import InstanceType
from firebolt.service.V2.types import EngineStatus
from firebolt.utils.exception import DatabaseNotFoundError

if TYPE_CHECKING:
    from firebolt.service.V2.engine import EngineService

logger = logging.getLogger(__name__)


@dataclass
class Engine(FireboltBaseModel):
    """
    A Firebolt engine. Responsible for performing work (queries, ingestion).
    """

    START_SQL: ClassVar[str] = 'START ENGINE "{}"'
    STOP_SQL: ClassVar[str] = 'STOP ENGINE "{}"'
    ALTER_PREFIX_SQL: ClassVar[str] = 'ALTER ENGINE "{}" '
    ALTER_PARAMETER_NAMES: ClassVar[Tuple] = (
        "NODES",
        "TYPE",
        "AUTO_STOP",
    )
    DROP_SQL: ClassVar[str] = 'DROP ENGINE "{}"'

    # Engine names can only contain alphanumeric characters and underscores
    _engine_name_re = re.compile(r"^[a-zA-Z0-9_]+$")

    _service: EngineService = field(repr=False, compare=False)

    name: str = field(metadata={"db_name": "engine_name"})
    region: str = field()
    spec: InstanceType = field(metadata={"db_name": "type"})
    scale: int = field(metadata={"db_name": "nodes"})
    current_status: EngineStatus = field(metadata={"db_name": "status"})
    _database_name: str = field(
        repr=False, metadata={"db_name": ("attached_to", "default_database")}
    )
    version: str = field()
    endpoint: str = field(metadata={"db_name": "url"})
    warmup: str = field()
    auto_stop: int = field()
    type: str = field(metadata={"db_name": "engine_type"})

    def __post_init__(self) -> None:
        # Specs are just strings for accounts v2
        if isinstance(self.spec, str) and self.spec:
            # Resolve engine specification
            self.spec = InstanceType(self.spec)
        if isinstance(self.current_status, str) and self.current_status:
            # Resolve engine status
            self.current_status = EngineStatus(self.current_status)

    @property
    def database(self) -> Optional[Database]:
        if self._database_name:
            try:
                return self._service.resource_manager.databases.get(self._database_name)
            except DatabaseNotFoundError:
                pass
        return None

    def refresh(self, name: Optional[str] = None) -> None:
        """Update attributes of the instance from Firebolt."""
        field_name_overrides = self._get_field_overrides()
        for field_name, value in self._service._get_dict(name or self.name).items():
            setattr(self, field_name_overrides.get(field_name, field_name), value)

        self.__post_init__()

    def attach_to_database(self, database: Union[Database, str]) -> None:
        """
        Attach this engine to a database.

        Args:
            database: Database to which the engine will be attached
        """
        self._service.attach_to_database(self, database)

    def get_connection(self) -> Connection:
        """Get a connection to the attached database for running queries.

        Returns:
            firebolt.db.connection.Connection: engine connection instance

        """
        return connect(
            database=self._database_name,  # type: ignore # already checked by decorator
            # we always have firebolt Auth as a client auth
            auth=self._service._connection._client.auth,  # type: ignore
            engine_name=self.name,
            account_name=self._service.resource_manager.account_name,
            api_endpoint=self._service.resource_manager.api_endpoint,
        )

    def _wait_for_start_stop(self) -> None:
        wait_timeout = 3600
        interval_seconds = 5
        timeout_time = time.time() + wait_timeout
        while self.current_status in (EngineStatus.STOPPING, EngineStatus.STARTING):
            logger.info(
                f"Engine {self.name} is currently "
                f"{self.current_status.value.lower()}, waiting"
            )
            time.sleep(interval_seconds)
            if time.time() > timeout_time:
                raise TimeoutError(
                    f"Excedeed timeout of {wait_timeout}s waiting for "
                    f"an engine in {self.current_status.value.lower()} state"
                )
            logger.info(".[!n]")
            self.refresh()

    def start(self) -> Engine:
        """
        Start an engine. If it's already started, do nothing.

        Returns:
            The updated engine instance.
        """

        self.refresh()
        self._wait_for_start_stop()
        if self.current_status == EngineStatus.RUNNING:
            logger.info(f"Engine {self.name} is already running.")
            return self
        if self.current_status in (EngineStatus.DROPPING, EngineStatus.REPAIRING):
            raise ValueError(
                f"Unable to start engine {self.name} because it's "
                f"in {self.current_status.value.lower()} state"
            )

        logger.info(f"Starting engine {self.name}")
        with self._service._connection.cursor() as c:
            c.execute(self.START_SQL.format(self.name))
        self.refresh()
        return self

    def stop(self) -> Engine:
        """Stop an engine. If it's already stopped, do nothing.

        Returns:
            The updated engine instance.
        """
        self.refresh()
        self._wait_for_start_stop()
        if self.current_status == EngineStatus.STOPPED:
            logger.info(f"Engine {self.name} is already stopped.")
            return self
        if self.current_status in (EngineStatus.DROPPING, EngineStatus.REPAIRING):
            raise ValueError(
                f"Unable to stop engine {self.name} because it's "
                f"in {self.current_status.value.lower()} state"
            )
        logger.info(f"Stopping engine {self.name}")
        with self._service._connection.cursor() as c:
            c.execute(self.STOP_SQL.format(self.name))
        self.refresh()
        return self

    def update(
        self,
        name: Optional[str] = None,
        engine_type: Optional[str] = None,
        scale: Optional[int] = None,
        spec: Union[InstanceType, str, None] = None,
        auto_stop: Optional[int] = None,
        warmup: Optional[str] = None,
    ) -> Engine:
        """
        Updates the engine and returns an updated version of the engine. If all
        parameters are set to None, old engine parameter values remain.
        """
        disallowed = [
            name
            for name, value in (("engine_type", engine_type), ("warmup", warmup))
            if value
        ]
        if disallowed:
            raise ValueError(
                f"Parameters {disallowed} are not supported for this account"
            )

        if not any(x is not None for x in (name, scale, spec, auto_stop)):
            # Nothing to be updated
            return self

        if name is not None and any(x is not None for x in (scale, spec, auto_stop)):
            raise ValueError("Cannot update name and other parameters at the same time")

        self.refresh()
        self._wait_for_start_stop()
        if self.current_status in (EngineStatus.DROPPING, EngineStatus.REPAIRING):
            raise ValueError(
                f"Unable to update engine {self.name} because it's "
                f"in {self.current_status.value.lower()} state"
            )

        sql = self.ALTER_PREFIX_SQL.format(self.name)
        parameters: List[Union[str, int]] = []
        if name is not None:
            if not self._engine_name_re.match(name):
                raise ValueError(
                    f"Engine name {name} is invalid, "
                    "it must only contain alphanumeric characters and underscores."
                )
            sql += f" RENAME TO {name}"
        else:
            sql += " SET "
            parameters = []
            for param, value in zip(
                self.ALTER_PARAMETER_NAMES,
                (scale, spec, auto_stop),
            ):
                if value is not None:
                    sql_part, new_params = self._service._format_engine_attribute_sql(
                        param, value
                    )
                    sql += sql_part
                    parameters.extend(new_params)

        with self._service._connection.cursor() as c:
            c.execute(sql, parameters)
        self.refresh(name)
        return self

    def delete(self) -> None:
        """Delete an engine."""
        self.refresh()
        if self.current_status in [EngineStatus.DROPPING, EngineStatus.DELETING]:
            return
        with self._service._connection.cursor() as c:
            c.execute(self.DROP_SQL.format(self.name))



from logging import getLogger
from typing import List, Optional, Union

from firebolt.model.V1.engine import Engine
from firebolt.service.V1.base import BaseService
from firebolt.service.V1.types import EngineOrder
from firebolt.utils.urls import (
    ACCOUNT_ENGINE_ID_BY_NAME_URL,
    ACCOUNT_ENGINE_URL,
    ACCOUNT_LIST_ENGINES_URL,
    ENGINES_BY_IDS_URL,
)
from firebolt.utils.util import prune_dict

logger = getLogger(__name__)


class EngineService(BaseService):
    def get(self, id_: str) -> Engine:
        """Get an engine from Firebolt by its ID."""

        response = self.client.get(
            url=ACCOUNT_ENGINE_URL.format(account_id=self.account_id, engine_id=id_),
        )
        engine_entry: dict = response.json()["engine"]
        return Engine.parse_obj_with_service(obj=engine_entry, engine_service=self)

    def get_by_ids(self, ids: List[str]) -> List[Engine]:
        """Get multiple engines from Firebolt by ID."""
        response = self.client.post(
            url=ENGINES_BY_IDS_URL,
            json={
                "engine_ids": [
                    {"account_id": self.account_id, "engine_id": engine_id}
                    for engine_id in ids
                ]
            },
        )
        return [
            Engine.parse_obj_with_service(obj=e, engine_service=self)
            for e in response.json()["engines"]
        ]

    def get_by_name(self, name: str) -> Engine:
        """Get an engine from Firebolt by its name."""

        response = self.client.get(
            url=ACCOUNT_ENGINE_ID_BY_NAME_URL.format(account_id=self.account_id),
            params={"engine_name": name},
        )
        engine_id = response.json()["engine_id"]["engine_id"]
        return self.get(id_=engine_id)

    def get_many(
        self,
        name_contains: Optional[str] = None,
        current_status_eq: Optional[str] = None,
        current_status_not_eq: Optional[str] = None,
        region_eq: Optional[str] = None,
        order_by: Optional[Union[str, EngineOrder]] = None,
    ) -> List[Engine]:
        """
        Get a list of engines on Firebolt.

        Args:
            name_contains: Filter for engines with a name containing this substring
            current_status_eq: Filter for engines with this status
            current_status_not_eq: Filter for engines that do not have this status
            region_eq: Filter for engines by region
            order_by: Method by which to order the results. See [EngineOrder]

        Returns:
            A list of engines matching the filters
        """

        if isinstance(order_by, str):
            order_by = EngineOrder[order_by].name

        if region_eq is not None:
            region_eq = self.resource_manager.regions.get_by_name(
                name=region_eq
            ).key.region_id

        response = self.client.get(
            url=ACCOUNT_LIST_ENGINES_URL.format(account_id=self.account_id),
            params=prune_dict(
                {
                    "page.first": 5000,
                    "filter.name_contains": name_contains,
                    "filter.current_status_eq": current_status_eq,
                    "filter.current_status_not_eq": current_status_not_eq,
                    "filter.compute_region_id_region_id_eq": region_eq,
                    "order_by": order_by,
                }
            ),
        )
        return [
            Engine.parse_obj_with_service(obj=e["node"], engine_service=self)
            for e in response.json()["edges"]
        ]



from __future__ import annotations

import logging
import re
from dataclasses import dataclass, fields
from enum import Enum
from functools import wraps
from types import TracebackType
from typing import Any, Callable, Dict, List, Optional, Tuple, Union

from httpx import URL, Response

from firebolt.common._types import (
    ColType,
    Column,
    RawColType,
    SetParameter,
    parse_type,
    parse_value,
)
from firebolt.utils.exception import (
    ConfigurationError,
    CursorClosedError,
    DataError,
    FireboltError,
    MethodNotAllowedInAsyncError,
    QueryNotRunError,
)
from firebolt.utils.util import Timer, fix_url_schema

logger = logging.getLogger(__name__)


JSON_OUTPUT_FORMAT = "JSON_Compact"


class CursorState(Enum):
    NONE = 1
    ERROR = 2
    DONE = 3
    CLOSED = 4


# Parameters that should be set using USE instead of SET
USE_PARAMETER_LIST = ["database", "engine"]
# parameters that can only be set by the backend
DISALLOWED_PARAMETER_LIST = ["output_format"]
# parameters that are set by the backend and should not be set by the user
IMMUTABLE_PARAMETER_LIST = USE_PARAMETER_LIST + DISALLOWED_PARAMETER_LIST

UPDATE_ENDPOINT_HEADER = "Firebolt-Update-Endpoint"
UPDATE_PARAMETERS_HEADER = "Firebolt-Update-Parameters"
RESET_SESSION_HEADER = "Firebolt-Reset-Session"


def _parse_update_parameters(parameter_header: str) -> Dict[str, str]:
    """Parse update parameters and set them as attributes."""
    # parse key1=value1,key2=value2 comma separated string into dict
    param_dict = dict(item.split("=") for item in parameter_header.split(","))
    # strip whitespace from keys and values
    param_dict = {key.strip(): value.strip() for key, value in param_dict.items()}
    return param_dict


def _parse_update_endpoint(
    new_engine_endpoint_header: str,
) -> Tuple[str, Dict[str, str]]:
    endpoint = URL(fix_url_schema(new_engine_endpoint_header))
    return fix_url_schema(endpoint.host), dict(endpoint.params)


def _raise_if_internal_set_parameter(parameter: SetParameter) -> None:
    """
    Check if parameter is internal and raise an error if it is.
    """
    if parameter.name in USE_PARAMETER_LIST:
        raise ConfigurationError(
            "Could not set parameter. "
            f"Set parameter '{parameter.name}' is not allowed. "
            f"Try again with 'USE {str(parameter.name).upper()}' instead of SET"
        )
    if parameter.name in DISALLOWED_PARAMETER_LIST:
        raise ConfigurationError(
            "Could not set parameter. "
            f"Set parameter '{parameter.name}' is not allowed. "
            "Try again with a different parameter name."
        )


@dataclass
class AsyncResponse:
    token: str
    message: str
    monitorSql: str


@dataclass
class Statistics:
    """
    Class for query execution statistics.
    """

    elapsed: float
    rows_read: int
    bytes_read: int
    time_before_execution: float
    time_to_execute: float
    scanned_bytes_cache: Optional[float] = None
    scanned_bytes_storage: Optional[float] = None

    def __post_init__(self) -> None:
        for field in fields(self):
            value = getattr(self, field.name)
            _type = eval(field.type)  # type: ignore

            # Unpack Optional
            if hasattr(_type, "__args__"):
                _type = _type.__args__[0]
            if value is not None and not isinstance(value, _type):
                # convert values to proper types
                setattr(self, field.name, _type(value))


RowSet = Tuple[
    int,
    Optional[List[Column]],
    Optional[Statistics],
    Optional[List[List[RawColType]]],
]


def check_not_closed(func: Callable) -> Callable:
    """(Decorator) ensure cursor is not closed before calling method."""

    @wraps(func)
    def inner(self: BaseCursor, *args: Any, **kwargs: Any) -> Any:
        if self.closed:
            raise CursorClosedError(method_name=func.__name__)
        return func(self, *args, **kwargs)

    return inner


def check_query_executed(func: Callable) -> Callable:
    """
    (Decorator) ensure that some query has been executed before
    calling cursor method.
    """

    @wraps(func)
    def inner(self: BaseCursor, *args: Any, **kwargs: Any) -> Any:
        if self._state == CursorState.NONE:
            raise QueryNotRunError(method_name=func.__name__)
        if self._query_token:
            # query_token is set only for async queries
            raise MethodNotAllowedInAsyncError(method_name=func.__name__)
        return func(self, *args, **kwargs)

    return inner


def async_not_allowed(func: Callable) -> Callable:
    """
    (Decorator) ensure that fetch methods are not called on async queries.
    """

    @wraps(func)
    def inner(self: BaseCursor, *args: Any, **kwargs: Any) -> Any:
        if self._query_token:
            # query_token is set only for async queries
            raise MethodNotAllowedInAsyncError(method_name=func.__name__)
        return func(self, *args, **kwargs)

    return inner


class BaseCursor:
    __slots__ = (
        "connection",
        "parameters",
        "_arraysize",
        "_client",
        "_state",
        "_descriptions",
        "_statistics",
        "_rowcount",
        "_rows",
        "_idx",
        "_row_sets",
        "_next_set_idx",
        "_set_parameters",
        "_query_id",
        "_query_token",
        "engine_url",
    )

    default_arraysize = 1

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        self._arraysize = self.default_arraysize
        # These fields initialized here for type annotations purpose
        self._rows: Optional[List[List[RawColType]]] = None
        self._descriptions: Optional[List[Column]] = None
        self._statistics: Optional[Statistics] = None
        self._row_sets: List[RowSet] = []
        # User-defined set parameters
        self._set_parameters: Dict[str, Any] = dict()
        # Server-side parameters (user can't change them)
        self.parameters: Dict[str, str] = dict()
        self.engine_url = ""
        self._rowcount = -1
        self._idx = 0
        self._next_set_idx = 0
        self._query_id = ""
        self._query_token = ""
        self._reset()

    @property
    def database(self) -> Optional[str]:
        return self.parameters.get("database")

    @database.setter
    def database(self, database: str) -> None:
        self.parameters["database"] = database

    @property  # type: ignore
    @check_not_closed
    def description(self) -> Optional[List[Column]]:
        """
        Provides information about a single result row of a query.

        Attributes:
            * ``name``
            * ``type_code``
            * ``display_size``
            * ``internal_size``
            * ``precision``
            * ``scale``
            * ``null_ok``
        """
        return self._descriptions

    @property  # type: ignore
    @check_not_closed
    def statistics(self) -> Optional[Statistics]:
        """Query execution statistics returned by the backend."""
        return self._statistics

    @property  # type: ignore
    @check_not_closed
    def rowcount(self) -> int:
        """The number of rows produced by last query."""
        return self._rowcount

    @property  # type: ignore
    @check_not_closed
    def query_id(self) -> str:
        """The query id of a query executed asynchronously."""
        return self._query_id

    @property
    def async_query_token(self) -> str:
        """The query token of a query executed asynchronously."""
        if not self._query_token:
            raise FireboltError(
                "No async query was executed or query was not an async."
            )
        return self._query_token

    @property
    def arraysize(self) -> int:
        """Default number of rows returned by fetchmany."""
        return self._arraysize

    @arraysize.setter
    def arraysize(self, value: int) -> None:
        if not isinstance(value, int):
            raise TypeError(
                "Invalid arraysize value type, expected int,"
                f" got {type(value).__name__}"
            )
        self._arraysize = value

    @property
    def closed(self) -> bool:
        """True if connection is closed, False otherwise."""
        return self._state == CursorState.CLOSED

    @check_not_closed
    @async_not_allowed
    @check_query_executed
    def nextset(self) -> Optional[bool]:
        """
        Skip to the next available set, discarding any remaining rows
        from the current set.
        Returns True if operation was successful;
        None if there are no more sets to retrive.
        """
        return self._pop_next_set()

    def _pop_next_set(self) -> Optional[bool]:
        """
        Same functionality as .nextset, but doesn't check that query has been executed.
        """
        if self._next_set_idx >= len(self._row_sets):
            return None
        (
            self._rowcount,
            self._descriptions,
            self._statistics,
            self._rows,
        ) = self._row_sets[self._next_set_idx]
        self._idx = 0
        self._next_set_idx += 1
        return True

    def flush_parameters(self) -> None:
        """Cleanup all previously set parameters"""
        self._set_parameters = dict()

    def _reset(self) -> None:
        """Clear all data stored from previous query."""
        self._state = CursorState.NONE
        self._rows = None
        self._descriptions = None
        self._statistics = None
        self._rowcount = -1
        self._idx = 0
        self._row_sets = []
        self._next_set_idx = 0
        self._query_id = ""
        self._query_token = ""

    def _update_set_parameters(self, parameters: Dict[str, Any]) -> None:
        # Split parameters into immutable and user parameters
        immutable_parameters = {
            key: value
            for key, value in parameters.items()
            if key in IMMUTABLE_PARAMETER_LIST
        }
        user_parameters = {
            key: value
            for key, value in parameters.items()
            if key not in IMMUTABLE_PARAMETER_LIST
        }

        self.parameters.update(immutable_parameters)

        self._set_parameters.update(user_parameters)

    def _update_server_parameters(self, parameters: Dict[str, Any]) -> None:
        for key, value in parameters.items():
            self.parameters[key] = value

    @staticmethod
    def _log_query(query: Union[str, SetParameter]) -> None:
        # Our CREATE EXTERNAL TABLE queries currently require credentials,
        # so we will skip logging those queries.
        # https://docs.firebolt.io/sql-reference/commands/create-external-table.html
        if isinstance(query, SetParameter) or not re.search(
            "aws_key_id|credentials", query, flags=re.IGNORECASE
        ):
            logger.debug(f"Running query: {query}")

    @property
    def engine_name(self) -> str:
        """
        Get the name of the engine that we're using.

        Args:
            engine_url (str): URL of the engine
        """
        if self.parameters.get("engine"):
            return self.parameters["engine"]
        return URL(self.engine_url).host.split(".")[0].replace("-", "_")

    def _parse_async_response(self, response: Response) -> None:
        """Handle async response from the server."""
        async_response = AsyncResponse(**response.json())
        self._query_token = async_response.token

    def _row_set_from_response(self, response: Response) -> RowSet:
        """Fetch information about executed query from http response."""

        # Empty response is returned for insert query
        if response.headers.get("content-length", "") == "0":
            return -1, None, None, None
        try:
            # Skip parsing floats to properly parse them later
            query_data = response.json(parse_float=str)
            rowcount = int(query_data["rows"])
            descriptions: Optional[List[Column]] = [
                Column(d["name"], parse_type(d["type"]), None, None, None, None, None)
                for d in query_data["meta"]
            ]
            if not descriptions:
                descriptions = None
            statistics = Statistics(**query_data["statistics"])
            # Parse data during fetch
            rows = query_data["data"]
            return rowcount, descriptions, statistics, rows
        except (KeyError, ValueError) as err:
            raise DataError(f"Invalid query data format: {str(err)}")

    def _append_row_set(
        self,
        row_set: RowSet,
    ) -> None:
        """Store information about executed query."""
        self._row_sets.append(row_set)
        if self._next_set_idx == 0:
            # Populate values for first set
            self._pop_next_set()

    def _parse_row(self, row: List[RawColType]) -> List[ColType]:
        """Parse a single data row based on query column types."""
        assert len(row) == len(self.description)
        return [
            parse_value(col, self.description[i].type_code) for i, col in enumerate(row)
        ]

    def _get_next_range(self, size: int) -> Tuple[int, int]:
        """
        Return range of next rows of size (if possible),
        and update _idx to point to the end of this range
        """

        if self._rows is None:
            # No elements to take
            raise DataError("no rows to fetch")

        left = self._idx
        right = min(self._idx + size, len(self._rows))
        self._idx = right
        return left, right

    _performance_log_message = (
        "[PERFORMANCE] Parsing query output into native Python types"
    )

    @check_not_closed
    @async_not_allowed
    @check_query_executed
    def fetchone(self) -> Optional[List[ColType]]:
        """Fetch the next row of a query result set."""
        left, right = self._get_next_range(1)
        if left == right:
            # We are out of elements
            return None
        assert self._rows is not None
        with Timer(self._performance_log_message):
            result = self._parse_row(self._rows[left])
        return result

    @check_not_closed
    @async_not_allowed
    @check_query_executed
    def fetchmany(self, size: Optional[int] = None) -> List[List[ColType]]:
        """
        Fetch the next set of rows of a query result;
        cursor.arraysize is default size.
        """
        size = size if size is not None else self.arraysize
        left, right = self._get_next_range(size)
        assert self._rows is not None
        rows = self._rows[left:right]
        with Timer(self._performance_log_message):
            result = [self._parse_row(row) for row in rows]
        return result

    @check_not_closed
    @async_not_allowed
    @check_query_executed
    def fetchall(self) -> List[List[ColType]]:
        """Fetch all remaining rows of a query result."""
        left, right = self._get_next_range(self.rowcount)
        assert self._rows is not None
        rows = self._rows[left:right]
        with Timer(self._performance_log_message):
            result = [self._parse_row(row) for row in rows]
        return result

    @check_not_closed
    def setinputsizes(self, sizes: List[int]) -> None:
        """Predefine memory areas for query parameters (does nothing)."""

    @check_not_closed
    def setoutputsize(self, size: int, column: Optional[int] = None) -> None:
        """Set a column buffer size for fetches of large columns (does nothing)."""

    def close(self) -> None:
        """Terminate an ongoing query (if any) and mark connection as closed."""
        self._state = CursorState.CLOSED
        self.connection._remove_cursor(self)  # type:ignore

    def __del__(self) -> None:
        self.close()

    def __exit__(
        self, exc_type: type, exc_val: Exception, exc_tb: TracebackType
    ) -> None:
        self.close()



from base64 import b64decode, b64encode, urlsafe_b64encode
from hashlib import sha256
from json import JSONDecodeError
from json import dump as json_dump
from json import load as json_load
from os import makedirs, path, urandom
from time import time
from typing import Optional

from appdirs import user_data_dir
from cryptography.fernet import Fernet, InvalidToken
from cryptography.hazmat.backends import default_backend  # type: ignore
from cryptography.hazmat.primitives import hashes  # type: ignore
from cryptography.hazmat.primitives.kdf.pbkdf2 import (
    PBKDF2HMAC,  # type: ignore
)

APPNAME = "firebolt"


def generate_salt() -> str:
    """Generate salt for FernetExcrypter.

    Returns:
        str: Generated salt
    """
    return b64encode(urandom(16)).decode("ascii")


def generate_file_name(username: str, password: str) -> str:
    """Generate unique file name based on username and password.

    Username and password values are not exposed.

    Args:
        username (str): Username
        password (str): Password

    Returns:
        str: File name 64 characters long

    """
    username_hash = sha256(username.encode("utf-8")).hexdigest()[:32]
    password_hash = sha256(password.encode("utf-8")).hexdigest()[:32]

    return f"{username_hash}{password_hash}.json"


class TokenSecureStorage:
    """File system storage for token.

    Token is encrypted using username and password.

    Args:
        username (str): Username
        password (str): Password
    """

    def __init__(self, username: str, password: str):
        self._data_dir = user_data_dir(appname=APPNAME)
        makedirs(self._data_dir, exist_ok=True)

        self._token_file = path.join(
            self._data_dir, generate_file_name(username, password)
        )

        self.salt = self._get_salt()
        self.encrypter = FernetEncrypter(self.salt, username, password)

    def _get_salt(self) -> str:
        """Get salt from the file if exists, or generate a new one.

        Returns:
            str: Salt
        """
        res = self._read_data_json()
        return res.get("salt", generate_salt())

    def _read_data_json(self) -> dict:
        """Read json token file.

        Returns:
            dict: JSON object
        """
        if not path.exists(self._token_file):
            return {}

        with open(self._token_file) as f:
            try:
                return json_load(f)
            except JSONDecodeError:
                return {}

    def get_cached_token(self) -> Optional[str]:
        """Get decrypted token.

        If token is not found, cannot be decrypted with username and password,
        or is expired - None will be returned.

        Returns:
            Optional[str]: Decrypted token or None
        """
        res = self._read_data_json()
        if "token" not in res:
            return None

        # Ignore expired tokens
        if "expiration" in res and res["expiration"] <= int(time()):
            return None

        return self.encrypter.decrypt(res["token"])

    def cache_token(self, token: str, expiration_ts: int) -> None:
        """Encrypt and store token in file system.

        Expiration timestamp is also stored with token in order to later
        be able to check if it's expired.

        Args:
            token (str): Token to store
            expiration_ts (int): Token expiration timestamp
        """
        token = self.encrypter.encrypt(token)

        with open(self._token_file, "w") as f:
            json_dump(
                {"token": token, "salt": self.salt, "expiration": expiration_ts}, f
            )


class FernetEncrypter:
    """PBDKF2HMAC based encrypter.

    Username and password combination is used as a key.

    Args:
        salt (str): Salt value for encryption
        username: Username for key
        password: Password for key
    """

    def __init__(self, salt: str, username: str, password: str):
        kdf = PBKDF2HMAC(
            algorithm=hashes.SHA256(),
            salt=b64decode(salt),
            length=32,
            iterations=39000,
            backend=default_backend(),
        )
        self.fernet = Fernet(
            urlsafe_b64encode(
                kdf.derive(bytes(f"{username}{password}", encoding="utf-8"))
            )
        )

    def encrypt(self, data: str) -> str:
        """Encrypt data string.

        Args:
            data (str): Data for encryption

        Returns:
            str: Encrypted data

        """
        return self.fernet.encrypt(bytes(data, encoding="utf-8")).decode("utf-8")

    def decrypt(self, data: str) -> Optional[str]:
        """Decrypt encrypted data.

        Args:
            data (str): Encrypted data

        Returns:
            Optional[str]: Decrypted data

        """
        try:
            return self.fernet.decrypt(bytes(data, encoding="utf-8")).decode("utf-8")
        except InvalidToken:
            return None



from __future__ import annotations

from datetime import datetime
from typing import Optional

from pydantic import BaseModel, Field

from firebolt.model.V1 import FireboltBaseModel


class BindingKey(BaseModel):
    account_id: str
    database_id: str
    engine_id: str


class Binding(FireboltBaseModel):
    """A binding between an engine and a database."""

    binding_key: BindingKey = Field(alias="id")
    is_default_engine: bool = Field(alias="engine_is_default")

    # optional
    current_status: Optional[str] = None
    health_status: Optional[str] = None
    create_time: Optional[datetime] = None
    create_actor: Optional[str] = None
    last_update_time: Optional[datetime] = None
    last_update_actor: Optional[str] = None
    desired_status: Optional[str] = None

    @property
    def database_id(self) -> str:
        return self.binding_key.database_id

    @property
    def engine_id(self) -> str:
        return self.binding_key.engine_id



from abc import ABCMeta, abstractmethod
from json import JSONDecodeError
from typing import Any, Dict, Optional

from httpx import URL
from httpx import AsyncClient as HttpxAsyncClient
from httpx import Client as HttpxClient
from httpx import HTTPStatusError, Request, RequestError, Response
from httpx import codes as HttpxCodes
from httpx._types import AuthTypes

from firebolt.client.auth import Auth
from firebolt.client.auth.base import AuthRequest
from firebolt.client.constants import (
    DEFAULT_API_URL,
    PROTOCOL_VERSION,
    PROTOCOL_VERSION_HEADER_NAME,
)
from firebolt.client.http_backend import (
    AsyncKeepaliveTransport,
    KeepaliveTransport,
)
from firebolt.utils.exception import (
    AccountNotFoundError,
    FireboltEngineError,
    InterfaceError,
)
from firebolt.utils.urls import (
    ACCOUNT_BY_NAME_URL_V1,
    ACCOUNT_ENGINE_ID_BY_NAME_URL,
    ACCOUNT_ENGINE_URL,
    ACCOUNT_ENGINE_URL_BY_DATABASE_NAME_V1,
    ACCOUNT_URL,
)
from firebolt.utils.util import (
    cached_property,
    fix_url_schema,
    get_auth_endpoint,
    merge_urls,
    mixin_for,
)

FireboltClientMixinBase = mixin_for(HttpxClient)  # type: Any


class FireboltClientMixin(FireboltClientMixinBase):
    """HttpxAsyncClient mixin with Firebolt authentication functionality."""

    def __init__(
        self,
        *args: Any,
        auth: Auth,
        account_name: Optional[str] = None,
        api_endpoint: str = DEFAULT_API_URL,
        **kwargs: Any,
    ):
        self.account_name = account_name
        self._api_endpoint = URL(fix_url_schema(api_endpoint))
        self._auth_endpoint = get_auth_endpoint(self._api_endpoint)
        super().__init__(*args, auth=auth, **kwargs)
        self._set_default_header(PROTOCOL_VERSION_HEADER_NAME, PROTOCOL_VERSION)

    def _set_default_header(self, key: str, value: str) -> None:
        if key not in self.headers:
            self.headers[key] = value

    def _build_auth(self, auth: Optional[AuthTypes]) -> Auth:
        """Create Auth object based on auth provided.

        Overrides ``httpx.Client._build_auth``

        Args:
            auth (AuthTypes): Provided auth

        Returns:
            Optional[Auth]: Auth object

        Raises:
            TypeError: Auth argument has unsupported type
        """
        if not (auth is None or isinstance(auth, Auth)):
            raise TypeError(f'Invalid "auth" argument: {auth!r}')
        assert auth is not None  # type check
        return auth

    def _merge_auth_request(self, request: Request) -> Request:
        if isinstance(request, AuthRequest):
            request.url = merge_urls(self._auth_endpoint, request.url)
            request._prepare(dict(request.headers))
        return request

    def _enforce_trailing_slash(self, url: URL) -> URL:
        """Don't automatically append trailing slash to a base url"""
        return url

    def clone(self) -> "Client":
        return self.__class__(
            auth=self.auth,
            account_name=self.account_name,
            base_url=self.base_url,
            api_endpoint=str(self._api_endpoint),
            timeout=self.timeout,
            headers=self.headers,
        )


class Client(FireboltClientMixin, HttpxClient, metaclass=ABCMeta):
    def __init__(self, *args: Any, **kwargs: Any):
        super().__init__(*args, **kwargs, transport=KeepaliveTransport())

    @property
    @abstractmethod
    def account_id(self) -> str:
        ...

    def _send_handling_redirects(
        self, request: Request, *args: Any, **kwargs: Any
    ) -> Response:
        return super()._send_handling_redirects(
            self._merge_auth_request(request), *args, **kwargs
        )


class ClientV2(Client):
    """An HTTP client, based on httpx.Client.

    Handles the authentication for Firebolt database.
    Authentication can be passed through auth keyword as a tuple or as a
    FireboltAuth instance
    """

    def __init__(
        self,
        *args: Any,
        auth: Auth,
        account_name: str,
        api_endpoint: str = DEFAULT_API_URL,
        **kwargs: Any,
    ):
        super().__init__(
            *args,
            auth=auth,
            account_name=account_name,
            api_endpoint=api_endpoint,
            **kwargs,
        )

    @property
    def account_id(self) -> str:
        """User account ID.

        If account_name was provided during Client construction, returns its ID;
        gets default account otherwise.

        Returns:
            str: Account ID

        Raises:
            AccountNotFoundError: No account found with provided name
        """
        return ""


class ClientV1(Client):
    """An HTTP client, based on httpx.Client.

    Handles the authentication for Firebolt database.
    Authentication can be passed through auth keyword as a tuple or as a
    FireboltAuth instance
    """

    def __init__(
        self,
        *args: Any,
        auth: Auth,
        account_name: Optional[str] = None,
        api_endpoint: str = DEFAULT_API_URL,
        **kwargs: Any,
    ):
        super().__init__(
            *args,
            auth=auth,
            account_name=account_name,
            api_endpoint=api_endpoint,
            **kwargs,
        )
        self._auth_endpoint = URL(fix_url_schema(api_endpoint))

    @cached_property
    def account_id(self) -> str:
        """User account ID.

        If account_name was provided during Client construction, returns its ID;
        gets default account otherwise.

        Returns:
            str: Account ID

        Raises:
            AccountNotFoundError: No account found with provided name
        """
        if self.account_name:
            response = self.get(
                url=ACCOUNT_BY_NAME_URL_V1, params={"account_name": self.account_name}
            )
            if response.status_code == HttpxCodes.NOT_FOUND:
                raise AccountNotFoundError(self.account_name)
            # process all other status codes
            response.raise_for_status()
            return response.json()["account_id"]

        # account_name isn't set, use the default account.
        return self.get(url=ACCOUNT_URL).json()["account"]["id"]

    def _get_database_default_engine_url(
        self,
        database: str,
    ) -> str:
        try:
            account_id = self.account_id
            response = self.get(
                url=ACCOUNT_ENGINE_URL_BY_DATABASE_NAME_V1.format(
                    account_id=account_id
                ),
                params={"database_name": database},
            )
            response.raise_for_status()
            return response.json()["engine_url"]
        except (
            JSONDecodeError,
            RequestError,
            RuntimeError,
            HTTPStatusError,
            KeyError,
        ) as e:
            raise InterfaceError(f"Unable to retrieve default engine endpoint: {e}.")

    def _resolve_engine_url(self, engine_name: str) -> str:
        account_id = self.account_id
        url = ACCOUNT_ENGINE_ID_BY_NAME_URL.format(account_id=account_id)
        try:
            response = self.get(
                url=url,
                params={"engine_name": engine_name},
            )
            response.raise_for_status()
            engine_id = response.json()["engine_id"]["engine_id"]
            url = ACCOUNT_ENGINE_URL.format(account_id=account_id, engine_id=engine_id)
            response = self.get(url=url)
            response.raise_for_status()
            return response.json()["engine"]["endpoint"]
        except HTTPStatusError as e:
            # Engine error would be 404.
            if e.response.status_code != 404:
                raise InterfaceError(
                    f"Error {e.__class__.__name__}: Unable to retrieve engine "
                    f"endpoint {url}."
                )
            # Once this is point is reached we've already authenticated with
            # the backend so it's safe to assume the cause of the error is
            # missing engine.
            raise FireboltEngineError(f"Firebolt engine {engine_name} does not exist.")
        except (JSONDecodeError, RequestError, RuntimeError) as e:
            raise InterfaceError(
                f"Error {e.__class__.__name__}: "
                f"Unable to retrieve engine endpoint {url}."
            )


class AsyncClient(FireboltClientMixin, HttpxAsyncClient, metaclass=ABCMeta):
    def __init__(self, *args: Any, **kwargs: Any):
        super().__init__(*args, **kwargs, transport=AsyncKeepaliveTransport())

    @property
    @abstractmethod
    async def account_id(self) -> str:
        ...

    async def _send_handling_redirects(
        self, request: Request, *args: Any, **kwargs: Any
    ) -> Response:
        return await super()._send_handling_redirects(
            self._merge_auth_request(request), *args, **kwargs
        )


class AsyncClientV2(AsyncClient):
    """An HTTP client, based on httpx.Client.

    Handles the authentication for Firebolt database.
    Authentication can be passed through auth keyword as a tuple or as a
    FireboltAuth instance
    """

    def __init__(
        self,
        *args: Any,
        auth: Auth,
        account_name: str,
        api_endpoint: str = DEFAULT_API_URL,
        **kwargs: Any,
    ):
        super().__init__(
            *args,
            auth=auth,
            account_name=account_name,
            api_endpoint=api_endpoint,
            **kwargs,
        )

    @property
    async def account_id(self) -> str:
        """User account ID.

        If account_name was provided during Client construction, returns its ID;
        gets default account otherwise.

        Returns:
            str: Account ID

        Raises:
            AccountNotFoundError: No account found with provided name
        """
        return ""


class AsyncClientV1(AsyncClient):
    """An HTTP client, based on httpx.Client.

    Handles the authentication for Firebolt database.
    Authentication can be passed through auth keyword as a tuple or as a
    FireboltAuth instance
    """

    def __init__(
        self,
        *args: Any,
        auth: Auth,
        account_name: Optional[str] = None,
        api_endpoint: str = DEFAULT_API_URL,
        **kwargs: Any,
    ):
        super().__init__(
            *args,
            auth=auth,
            account_name=account_name,
            api_endpoint=api_endpoint,
            **kwargs,
        )
        self.account_id_cache: Dict[str, str] = {}
        self._auth_endpoint = URL(fix_url_schema(api_endpoint))

    @property
    async def account_id(self) -> str:
        """User account ID.

        If account_name was provided during Client construction, returns its ID;
        gets default account otherwise.

        Returns:
            str: Account ID

        Raises:
            AccountNotFoundError: No account found with provided name
        """
        # manual caching to avoid async_cached_property issues
        if self.account_name in self.account_id_cache:
            return self.account_id_cache[self.account_name]

        if self.account_name:
            response = await self.get(
                url=ACCOUNT_BY_NAME_URL_V1, params={"account_name": self.account_name}
            )
            if response.status_code == HttpxCodes.NOT_FOUND:
                raise AccountNotFoundError(self.account_name)
            # process all other status codes
            response.raise_for_status()
            account_id = response.json()["account_id"]
        else:
            # account_name isn't set; use the default account.
            account_response = (await self.get(url=ACCOUNT_URL)).json()["account"]
            account_id = account_response["id"]
            self.account_name = account_response["name"]
            assert self.account_name is not None  # type check
        # cache for future use
        self.account_id_cache[self.account_name] = account_id
        return account_id

    async def _get_database_default_engine_url(
        self,
        database: str,
    ) -> str:
        try:
            account_id = await self.account_id
            response = await self.get(
                url=ACCOUNT_ENGINE_URL_BY_DATABASE_NAME_V1.format(
                    account_id=account_id
                ),
                params={"database_name": database},
            )
            response.raise_for_status()
            return response.json()["engine_url"]
        except (
            JSONDecodeError,
            RequestError,
            RuntimeError,
            HTTPStatusError,
            KeyError,
        ) as e:
            raise InterfaceError(f"Unable to retrieve default engine endpoint: {e}.")

    async def _resolve_engine_url(self, engine_name: str) -> str:
        account_id = await self.account_id
        url = ACCOUNT_ENGINE_ID_BY_NAME_URL.format(account_id=account_id)
        try:
            response = await self.get(
                url=url,
                params={"engine_name": engine_name},
            )
            response.raise_for_status()
            engine_id = response.json()["engine_id"]["engine_id"]
            url = ACCOUNT_ENGINE_URL.format(account_id=account_id, engine_id=engine_id)
            response = await self.get(url=url)
            response.raise_for_status()
            return response.json()["engine"]["endpoint"]
        except HTTPStatusError as e:
            # Engine error would be 404.
            if e.response.status_code != 404:
                raise InterfaceError(
                    f"Error {e.__class__.__name__}: Unable to retrieve engine "
                    f"endpoint {url}."
                )
            # Once this is point is reached we've already authenticated with
            # the backend so it's safe to assume the cause of the error is
            # missing engine.
            raise FireboltEngineError(f"Firebolt engine {engine_name} does not exist.")
        except (JSONDecodeError, RequestError, RuntimeError) as e:
            raise InterfaceError(
                f"Error {e.__class__.__name__}: "
                f"Unable to retrieve engine endpoint {url}."
            )



import socket
from typing import Any

try:
    from httpcore.backends.auto import AutoBackend  # type: ignore
    from httpcore.backends.sync import SyncBackend  # type: ignore
except ImportError:
    from httpcore._backends.auto import AutoBackend  # type: ignore
    from httpcore._backends.sync import SyncBackend  # type: ignore

from httpx import AsyncHTTPTransport, HTTPTransport

from firebolt.common.constants import KEEPALIVE_FLAG, KEEPIDLE_RATE


def override_stream(stream):  # type: ignore [no-untyped-def]
    keepidle = getattr(socket, "TCP_KEEPIDLE", 0x10)  # 0x10 is TCP_KEEPALIVE on mac

    sock = (
        stream.get_extra_info("socket")
        if hasattr(stream, "get_extra_info")
        else stream.sock
    )

    # Enable keepalive
    sock.setsockopt(socket.SOL_SOCKET, socket.SO_KEEPALIVE, KEEPALIVE_FLAG)
    # Set keepalive to 60 seconds
    sock.setsockopt(socket.IPPROTO_TCP, keepidle, KEEPIDLE_RATE)
    return stream


class AsyncOverriddenHttpBackend(AutoBackend):
    """
    `OverriddenHttpBackend` is a short-term solution for the TCP
    connection idle timeout issue described in the following article:
    https://docs.aws.amazon.com/elasticloadbalancing/latest/network/network-load-balancers.html#connection-idle-timeout
    Since httpx creates a connection right before executing a request, the
    backend must be overridden to set the socket to `KEEPALIVE`
    and `KEEPIDLE` settings.
    """

    async def connect_tcp(self, *args, **kwargs):  # type: ignore
        stream = await super().connect_tcp(*args, **kwargs)
        return override_stream(stream)

    async def open_tcp_stream(self, *args, **kwargs):  # type: ignore
        stream = await super().open_tcp_stream(*args, **kwargs)
        return override_stream(stream)


class OverriddenHttpBackend(SyncBackend):
    """
    `OverriddenHttpBackend` is a short-term solution for the TCP
    connection idle timeout issue described in the following article:
    https://docs.aws.amazon.com/elasticloadbalancing/latest/network/network-load-balancers.html#connection-idle-timeout
    Since httpx creates a connection right before executing a request, the
    backend must be overridden to set the socket to `KEEPALIVE`
    and `KEEPIDLE` settings.
    """

    def connect_tcp(self, *args, **kwargs):  # type: ignore
        stream = super().connect_tcp(*args, **kwargs)
        return override_stream(stream)

    def open_tcp_stream(self, *args, **kwargs):  # type: ignore
        stream = super().open_tcp_stream(*args, **kwargs)
        return override_stream(stream)


class AsyncKeepaliveTransport(AsyncHTTPTransport):
    def __init__(self, *args: Any, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)
        if hasattr(self._pool, "_network_backend"):
            self._pool._network_backend = AsyncOverriddenHttpBackend()  # type: ignore
        if hasattr(self._pool, "_backend"):
            self._pool._backend = AsyncOverriddenHttpBackend()  # type: ignore


class KeepaliveTransport(HTTPTransport):
    def __init__(self, *args: Any, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)
        if hasattr(self._pool, "_network_backend"):
            self._pool._network_backend = OverriddenHttpBackend()  # type: ignore
        if hasattr(self._pool, "_backend"):
            self._pool._backend = OverriddenHttpBackend()  # type: ignore



from typing import TYPE_CHECKING, Optional

from firebolt.client import ClientV1 as Client

if TYPE_CHECKING:
    from firebolt.service.manager import ResourceManager


class BaseService:
    def __init__(self, resource_manager: "ResourceManager"):
        self.resource_manager = resource_manager

    @property
    def client(self) -> Client:
        return self.resource_manager._client

    @property
    def account_id(self) -> str:
        return self.resource_manager.account_id

    @property
    def default_region_setting(self) -> Optional[str]:
        return self.resource_manager.default_region



import logging
from typing import List, Optional, Union

from firebolt.model.V2.database import Database
from firebolt.model.V2.engine import Engine
from firebolt.service.V2.base import BaseService
from firebolt.utils.exception import DatabaseNotFoundError

logger = logging.getLogger(__name__)


class DatabaseService(BaseService):
    DB_FIELDS = (
        "catalog_name",
        "description",
        "created",
        "catalog_owner",
    )
    GET_SQL = f"SELECT {', '.join(DB_FIELDS)} FROM information_schema.catalogs"
    GET_BY_NAME_SQL = GET_SQL + " WHERE catalog_name=?"
    GET_WHERE_SQL = " WHERE "

    CREATE_PREFIX_SQL = 'CREATE DATABASE {}"{}"'
    IF_NOT_EXISTS_SQL = "IF NOT EXISTS "

    DISALLOWED_ACCOUNT_V2_PARAMETERS = [
        "attached_engine_name_eq",
        "attached_engine_name_contains",
    ]

    def _get_dict(self, name: str) -> dict:
        with self._connection.cursor() as c:
            count = c.execute(self.GET_BY_NAME_SQL, (name,))
            if count == 0:
                raise DatabaseNotFoundError(name)
            return {
                column.name: value for column, value in zip(c.description, c.fetchone())
            }

    def get(self, name: str) -> Database:
        """Get a Database from Firebolt by its name."""
        return Database._from_dict(self._get_dict(name), self)

    def get_by_name(self, name: str) -> Database:
        return self.get(name)

    def get_many(
        self,
        name_contains: Optional[str] = None,
        attached_engine_name_eq: Optional[str] = None,
        attached_engine_name_contains: Optional[str] = None,
        region_eq: Optional[str] = None,
    ) -> List[Database]:
        """
        Get a list of databases on Firebolt.

        Args:
            name_contains: Filter for databases with a name containing this substring
            attached_engine_name_eq: Filter for databases by an exact engine name
            attached_engine_name_contains: Filter for databases by engines with a
                name containing this substring
            region_eq: Filter for database by region

        Returns:
            A list of databases matching the filters
        """
        sql = self.GET_SQL
        parameters = []
        disallowed_parameters = [
            name
            for name, value in (
                ("attached_engine_name_eq", attached_engine_name_eq),
                ("attached_engine_name_contains", attached_engine_name_contains),
                ("region_eq", region_eq),
            )
            if value
        ]
        if disallowed_parameters:
            raise ValueError(
                f"Parameters {disallowed_parameters} are not supported for this account"
            )

        if name_contains:
            sql += " WHERE catalog_name like ?"
            parameters.append(f"%{name_contains}%")

        with self._connection.cursor() as c:
            c.execute(sql, parameters)
            dicts = [
                {column.name: value for column, value in zip(c.description, row)}
                for row in c.fetchall()
            ]
            return [Database._from_dict(_dict, self) for _dict in dicts]

    def create(
        self,
        name: str,
        region: Optional[str] = None,
        attached_engines: Union[List[str], List[Engine], None] = None,
        description: Optional[str] = None,
        fail_if_exists: bool = True,
    ) -> Database:
        """
        Create a new Database on Firebolt.

        Args:
            name: Name of the database
            region: Region name in which to create the database
            attached_engines: List of engines to attach to the database
            description: Description of the database
            fail_if_exists: Fail is a database with provided name already exists

        Returns:
            The newly created database
        """

        logger.info(f"Creating database {name}")

        disallowed_parameters = [
            name
            for name, value in (
                ("region", region),
                (attached_engines, attached_engines),
            )
            if value is not None
        ]
        if disallowed_parameters:
            raise ValueError(
                f"Parameters {disallowed_parameters} are not supported for this account"
            )

        sql = self.CREATE_PREFIX_SQL.format(
            ("" if fail_if_exists else self.IF_NOT_EXISTS_SQL), name
        )
        parameters = []
        if description:
            sql += " WITH DESCRIPTION = ? "
            parameters.append(description)

        with self._connection.cursor() as c:
            c.execute(sql, parameters)
        return self.get(name)



import logging
from typing import List, Optional, Union

from firebolt.model.V1 import FireboltBaseModel
from firebolt.model.V1.database import Database
from firebolt.service.V1.base import BaseService
from firebolt.service.V1.types import DatabaseOrder
from firebolt.utils.urls import (
    ACCOUNT_DATABASE_BY_NAME_URL,
    ACCOUNT_DATABASE_URL,
    ACCOUNT_DATABASES_URL,
)
from firebolt.utils.util import prune_dict

logger = logging.getLogger(__name__)


class DatabaseService(BaseService):
    def get(self, id_: str) -> Database:
        """Get a Database from Firebolt by its ID."""

        response = self.client.get(
            url=ACCOUNT_DATABASE_URL.format(account_id=self.account_id, database_id=id_)
        )
        return Database.parse_obj_with_service(
            obj=response.json()["database"], database_service=self
        )

    def get_by_name(self, name: str) -> Database:
        """Get a database from Firebolt by its name."""

        database_id = self.get_id_by_name(name=name)
        return self.get(id_=database_id)

    def get_id_by_name(self, name: str) -> str:
        """Get a database ID from Firebolt by its name."""

        response = self.client.get(
            url=ACCOUNT_DATABASE_BY_NAME_URL.format(account_id=self.account_id),
            params={"database_name": name},
        )
        database_id = response.json()["database_id"]["database_id"]
        return database_id

    def get_many(
        self,
        name_contains: Optional[str] = None,
        attached_engine_name_eq: Optional[str] = None,
        attached_engine_name_contains: Optional[str] = None,
        order_by: Optional[Union[str, DatabaseOrder]] = None,
    ) -> List[Database]:
        """
        Get a list of databases on Firebolt.

        Args:
            name_contains: Filter for databases with a name containing this substring
            attached_engine_name_eq: Filter for databases by an exact engine name
            attached_engine_name_contains: Filter for databases by engines with a
                name containing this substring
            order_by: Method by which to order the results.
                See :py:class:`firebolt.service.types.DatabaseOrder`

        Returns:
            A list of databases matching the filters
        """

        if isinstance(order_by, str):
            order_by = DatabaseOrder[order_by].name

        params = {
            "page.first": "1000",
            "order_by": order_by,
            "filter.name_contains": name_contains,
            "filter.attached_engine_name_eq": attached_engine_name_eq,
            "filter.attached_engine_name_contains": attached_engine_name_contains,
        }

        response = self.client.get(
            url=ACCOUNT_DATABASES_URL.format(account_id=self.account_id),
            params=prune_dict(params),
        )

        return [
            Database.parse_obj_with_service(obj=d["node"], database_service=self)
            for d in response.json()["edges"]
        ]

    def create(
        self, name: str, region: Optional[str] = None, description: Optional[str] = None
    ) -> Database:
        """
        Create a new Database on Firebolt.

        Args:
            name: Name of the database
            region: Region name in which to create the database

        Returns:
            The newly created database
        """

        class _DatabaseCreateRequest(FireboltBaseModel):
            """Helper model for sending database creation requests."""

            account_id: str
            database: Database

        if region is None:
            region_key = self.resource_manager.regions.default_region.key
        else:
            region_key = self.resource_manager.regions.get_by_name(name=region).key
        database = Database(
            name=name,
            compute_region_key=region_key,  # type: ignore[call-arg]
            description=description,
        )

        logger.info(f"Creating Database (name={name})")
        response = self.client.post(
            url=ACCOUNT_DATABASES_URL.format(account_id=self.account_id),
            headers={"Content-type": "application/json"},
            json=_DatabaseCreateRequest(
                account_id=self.account_id,
                database=database,
            ).jsonable_dict(by_alias=True),
        )
        return Database.parse_obj_with_service(
            obj=response.json()["database"], database_service=self
        )



from typing import TYPE_CHECKING

from firebolt.db import Connection

if TYPE_CHECKING:
    from firebolt.service.manager import ResourceManager


class BaseService:
    def __init__(self, resource_manager: "ResourceManager"):
        self.resource_manager = resource_manager

    @property
    def _connection(self) -> Connection:
        return self.resource_manager._connection



import logging
import os
from dataclasses import dataclass, field
from typing import Any, Callable

from firebolt.client.auth import Auth, ClientCredentials

logger = logging.getLogger(__name__)

CLIENT_ID_ENV = "FIREBOLT_CLIENT_ID"
CLIENT_SECRET_ENV = "FIREBOLT_CLIENT_SECRET"
ACCOUNT_ENV = "FIREBOLT_ACCOUNT"
SERVER_ENV = "FIREBOLT_SERVER"
DEFAULT_REGION_ENV = "FIREBOLT_DEFAULT_REGION"


def from_env(var_name: str, default: Any = None) -> Callable:
    def inner() -> Any:
        return os.environ.get(var_name, default)

    return inner


def auth_from_env() -> Auth:
    client_id = os.environ.get(CLIENT_ID_ENV, None)
    client_secret = os.environ.get(CLIENT_SECRET_ENV, None)
    if client_id and client_secret:
        return ClientCredentials(client_id, client_secret)
    raise ValueError("Auth not provided")


@dataclass
class Settings:
    """Settings for Firebolt SDK.

    Attributes:
        user (Optional[str]): User name
        password (Optional[str]): User password
        access_token (Optional[str]): Access token to use for authentication
            Mutually exclusive with user and password
        account_name (Optional[str]): Account name
            Default user account is used if none provided
        server (Optional[str]): Environment api endpoint (Advanced)
            Default api endpoint is used if none provided
        default_region (str): Default region for provisioning
    """

    auth: Auth = field(default_factory=auth_from_env)

    account_name: str = field(default_factory=from_env(ACCOUNT_ENV))
    server: str = field(default_factory=from_env(SERVER_ENV))
    default_region: str = field(default_factory=from_env(DEFAULT_REGION_ENV))



import json
from dataclasses import dataclass, field, fields
from typing import ClassVar, Dict, Iterable, Optional, Type, TypeVar

from firebolt.service.V2.base import BaseService

Model = TypeVar("Model", bound="FireboltBaseModel")


@dataclass
class FireboltBaseModel:
    _service: BaseService = field(repr=False, compare=False)

    @classmethod
    def _get_field_overrides(cls) -> Dict[str, str]:
        """Create a mapping of db field name to class name where they are different."""
        return {
            override: f.name
            for f in fields(cls)
            if "db_name" in f.metadata
            for override in (
                [f.metadata["db_name"]]
                if isinstance(f.metadata["db_name"], str)
                else f.metadata["db_name"]
            )
        }

    @classmethod
    def _from_dict(
        cls: Type[Model], data: dict, service: Optional[BaseService] = None
    ) -> Model:
        data["_service"] = service
        field_name_overrides = cls._get_field_overrides()
        return cls(**{field_name_overrides.get(k, k): v for k, v in data.items()})



AUTH_SERVICE_ACCOUNT_URL = "/oauth/token"

DATABASES_URL = "/core/v1/account/databases"

ENGINES_URL = "/core/v1/account/engines"
ENGINES_BY_IDS_URL = "/core/v1/engines:getByIds"

ACCOUNT_DATABASES_URL = "/core/v1/accounts/{account_id}/databases"
ACCOUNT_DATABASE_URL = "/core/v1/accounts/{account_id}/databases/{database_id}"
ACCOUNT_DATABASE_BINDING_URL = ACCOUNT_DATABASE_URL + "/bindings/{engine_id}"
ACCOUNT_DATABASE_BY_NAME_URL = ACCOUNT_DATABASES_URL + ":getIdByName"

ACCOUNT_INSTANCE_TYPES_URL = "/aws/v2/accounts/{account_id}/instanceTypes"

PROVIDERS_URL = "/compute/v1/providers"
REGIONS_URL = "/compute/v1/regions"

GATEWAY_HOST_BY_ACCOUNT_NAME = "/web/v3/account/{account_name}/engineUrl"

# V1 URLS

ACCOUNT_ENGINE_URL = "/core/v1/accounts/{account_id}/engines/{engine_id}"
ACCOUNT_ENGINE_START_URL = ACCOUNT_ENGINE_URL + ":start"
ACCOUNT_ENGINE_RESTART_URL = ACCOUNT_ENGINE_URL + ":restart"
ACCOUNT_ENGINE_STOP_URL = ACCOUNT_ENGINE_URL + ":stop"
ACCOUNT_LIST_ENGINES_URL = "/core/v1/accounts/{account_id}/engines"
ACCOUNT_ENGINE_ID_BY_NAME_URL = ACCOUNT_LIST_ENGINES_URL + ":getIdByName"
ACCOUNT_ENGINE_REVISION_URL = ACCOUNT_ENGINE_URL + "/engineRevisions/{revision_id}"

AUTH_URL = "/auth/v1/login"
ACCOUNT_ENGINE_URL_BY_DATABASE_NAME_V1 = (
    ACCOUNT_LIST_ENGINES_URL + ":getURLByDatabaseName"
)
ACCOUNT_URL = "/iam/v2/account"
ACCOUNT_BY_NAME_URL_V1 = "/iam/v2/accounts:getIdByName"
ACCOUNT_BINDINGS_URL = "/core/v1/accounts/{account_id}/bindings"



import logging
from functools import lru_cache
from os import environ
from time import time
from types import TracebackType
from typing import (
    TYPE_CHECKING,
    Callable,
    Dict,
    Optional,
    Tuple,
    Type,
    TypeVar,
)
from urllib.parse import parse_qs, urljoin, urlparse

from httpx import URL, Response, codes

from firebolt.utils.exception import (
    ConfigurationError,
    FireboltStructuredError,
)

T = TypeVar("T")
logger = logging.getLogger(__name__)


def cached_property(func: Callable[..., T]) -> T:
    """cached_property implementation for 3.7 backward compatibility.

    Args:
        func (Callable): Property getter

    Returns:
        T: Property of type, returned by getter
    """
    return property(lru_cache()(func))  # type: ignore


def prune_dict(d: dict) -> dict:
    """Prune items from dictionaries where value is None.

    Args:
        d (dict): Dict to prune

    Returns:
        dict: Pruned dict
    """
    return {k: v for k, v in d.items() if v is not None}


TMix = TypeVar("TMix")


def mixin_for(baseclass: Type[TMix]) -> Type[TMix]:
    """Define mixin with baseclass typehint.

    Should be used as a mixin base class to fix typehints.

    Args:
        baseclass (Type[TMix]): Class which mixin will be made for

    Returns:
        Type[TMix]: Mixin type to inherit from

    Examples:
        ```
        class ReadonlyMixin(mixin_for(BaseClass))):
            ...
        ```

    """
    if TYPE_CHECKING:
        return baseclass
    return object


def fix_url_schema(url: str) -> str:
    """Add schema to URL if it's missing.

    Args:
        url (str): URL to check

    Returns:
        str: URL with schema present

    """
    return url if url.startswith("http") else f"https://{url}"


def get_auth_endpoint(api_endpoint: URL) -> URL:
    """Create auth endpoint from api endpoint.

    Args:
        api_endpoint (URL): provided API endpoint

    Returns:
        URL: authentication endpoint
    """
    return api_endpoint.copy_with(
        host=".".join(["id"] + api_endpoint.host.split(".")[1:])
    )


def get_internal_error_code(response: Response) -> Optional[int]:
    """Get internal error code from response.

    Args:
        response (Response): HTTP response

    Returns:
        Optional[int]: Internal error code
    """
    # Internal server error usually hides the real error code in the response body
    if response.status_code == codes.INTERNAL_SERVER_ERROR:
        try:
            # Example response:
            # Received error from remote server
            # /core/v1/accounts/<account_num>/engines:getIdByName?engine_name=
            # <engine_name>. HTTP status code: 401 Unauthorized, body: failed to
            # verify JWT token: failed to verify jwt: "exp" not satisfied\n'
            error = int(response.text.split("HTTP status code: ")[1].split(" ")[0])
            body = (
                response.text.split("body: ")[1] if "body: " in response.text else None
            )
            logger.debug(
                f"Detected an internal server error with code: {error}, body: {body}"
            )
            return error
        except (IndexError, ValueError):
            return None
    return None


def merge_urls(base: URL, merge: URL) -> URL:
    """Merge a base and merge urls.

    If merge is not a relative url, do nothing

    Args:
        base (URL): Base URL to merge to
        merge (URL): URL to merge

    Returns:
        URL: Resulting URL
    """
    if merge.is_relative_url:
        merge_raw_path = base.raw_path + merge.raw_path.lstrip(b"/")
        return base.copy_with(raw_path=merge_raw_path)
    return merge


def validate_engine_name_and_url_v1(
    engine_name: Optional[str], engine_url: Optional[str]
) -> None:
    if engine_name and engine_url:
        raise ConfigurationError(
            "Both engine_name and engine_url are provided. Provide only one to connect."
        )


def _print_error_body(resp: Response) -> None:
    """log error body if it exists, since it's not always logged by default"""
    try:
        if (
            codes.is_error(resp.status_code)
            and "Content-Length" in resp.headers
            and int(resp.headers["Content-Length"]) > 0
        ):
            logger.error(f"Something went wrong: {resp.read().decode('utf-8')}")
    except Exception:
        pass


def raise_errors_from_body(resp: Response) -> None:
    """
    Process error in response body. Only raise errors if the json body
    can be parsed and contains errors. Otherwise, let the rest of the code
    handle the error.

    Args:
        resp (Response): HTTP response
    """
    to_raise = None
    try:
        decoded = resp.json()
        if "errors" in decoded and len(decoded["errors"]) > 0:
            # Raise later to avoid catching it in the except block
            to_raise = FireboltStructuredError(decoded)

    except Exception:
        # If we can't parse the body, let the rest of the code handle it
        # we can't raise an exception here because it would mask the original error
        pass

    if to_raise:
        raise to_raise


class Timer:
    def __init__(self, message: str = ""):
        self._message = message

    def __enter__(self) -> "Timer":
        self._start_time: float = time()
        return self

    def __exit__(
        self,
        exc_type: Type[BaseException],
        exc_val: BaseException,
        exc_tb: TracebackType,
    ) -> None:
        self.elapsed_time: str = "{:.2f}".format(round((time() - self._start_time), 2))
        if (
            environ.get("FIREBOLT_SDK_PERFORMANCE_DEBUG", "0") == "1"
            and self._message != ""
        ):
            log_message = self._message + self.elapsed_time + "s"
            logger.debug(log_message)


def parse_url_and_params(url: str) -> Tuple[str, Dict[str, str]]:
    """Extract URL and query parameters separately from a URL."""
    url = fix_url_schema(url)
    parsed_url = urlparse(url)
    query_params = parse_qs(parsed_url.query)
    # This strips query parameters from the URL by joining base URL and path
    # skipping the query parameters.
    result_url = urljoin(url, parsed_url.path)
    # parse_qs returns a dictionary with values as lists.
    # We want the last value in the list.
    query_params_dict = {}
    for key, values in query_params.items():
        # Multiple values for the same key are not expected
        if len(values) > 1:
            raise ValueError(f"Multiple values found for key '{key}'")
        query_params_dict[key] = values[0]
    return result_url, query_params_dict



import logging
from typing import List, Optional

from firebolt.model.V1.binding import Binding, BindingKey
from firebolt.model.V1.database import Database
from firebolt.model.V1.engine import Engine
from firebolt.service.V1.base import BaseService
from firebolt.service.V1.database import DatabaseService
from firebolt.service.V1.engine import EngineService
from firebolt.utils.exception import AlreadyBoundError
from firebolt.utils.urls import (
    ACCOUNT_BINDINGS_URL,
    ACCOUNT_DATABASE_BINDING_URL,
)
from firebolt.utils.util import prune_dict

logger = logging.getLogger(__name__)


class BindingService(BaseService):
    def get_by_key(self, binding_key: BindingKey) -> Binding:
        """Get a binding by its BindingKey"""
        response = self.client.get(
            url=ACCOUNT_DATABASE_BINDING_URL.format(
                account_id=binding_key.account_id,
                database_id=binding_key.database_id,
                engine_id=binding_key.engine_id,
            )
        )
        binding: dict = response.json()["binding"]
        return Binding.parse_model(binding)

    def get_many(
        self,
        database_id: Optional[str] = None,
        engine_id: Optional[str] = None,
        is_system_database: Optional[bool] = None,
    ) -> List[Binding]:
        """
        List bindings on Firebolt, optionally filtering by database and engine.

        Args:
            database_id:
                Return bindings matching the database_id.
                If None, match any databases.
            engine_id:
                Return bindings matching the engine_id.
                If None, match any engines.
            is_system_database:
                If True, return only system databases.
                If False, return only non-system databases.
                If None, do not filter on this parameter.

        Returns:
            List of bindings matching the filter parameters
        """

        response = self.client.get(
            url=ACCOUNT_BINDINGS_URL.format(account_id=self.account_id),
            params=prune_dict(
                {
                    "page.first": 5000,  # FUTURE: pagination support w/ generator
                    "filter.id_database_id_eq": database_id,
                    "filter.id_engine_id_eq": engine_id,
                    "filter.is_system_database_eq": is_system_database,
                }
            ),
        )
        return [Binding.parse_model(i["node"]) for i in response.json()["edges"]]

    def get_database_bound_to_engine(self, engine: Engine) -> Optional[Database]:
        """Get the database to which an engine is bound, if any."""
        try:
            binding = self.get_many(engine_id=engine.engine_id)[0]
        except IndexError:
            return None
        try:
            assert isinstance(
                self.resource_manager.databases, DatabaseService
            ), "Expected DatabaseService V1"
            return self.resource_manager.databases.get(id_=binding.database_id)
        except (KeyError, IndexError):
            return None

    def get_engines_bound_to_database(self, database: Database) -> List[Engine]:
        """Get a list of engines that are bound to a database."""

        bindings = self.get_many(database_id=database.database_id)
        if not bindings:
            return []
        assert isinstance(
            self.resource_manager.engines, EngineService
        ), "Expected EngineService V1"
        return self.resource_manager.engines.get_by_ids(
            ids=[b.engine_id for b in bindings]
        )

    def create(
        self, engine: Engine, database: Database, is_default_engine: bool
    ) -> Binding:
        """
        Create a new binding between an engine and a database.

        Args:
            engine: Engine to bind.
            database: Database to bind.
            is_default_engine:
                Whether this engine should be used as default for this database.
                Only one engine can be set as default for a single database.
                This will overwrite any existing default.

        Returns:
            New binding between the engine and database.
        """

        existing_database = self.get_database_bound_to_engine(engine=engine)
        if existing_database is not None:
            raise AlreadyBoundError(
                f"The engine {engine.name} is already bound "
                f"to {existing_database.name}!"
            )

        logger.info(
            f"Attaching Engine (engine_id={engine.engine_id}, name={engine.name}) "
            f"to Database (database_id={database.database_id}, "
            f"name={database.name})"
        )
        assert database.database_id is not None, "Database must have database_id"
        binding = Binding(
            binding_key=BindingKey(  # type: ignore[call-arg]
                account_id=self.account_id,
                database_id=database.database_id,
                engine_id=engine.engine_id,
            ),
            is_default_engine=is_default_engine,  # type: ignore[call-arg]
        )

        response = self.client.post(
            url=ACCOUNT_DATABASE_BINDING_URL.format(
                account_id=self.account_id,
                database_id=database.database_id,
                engine_id=engine.engine_id,
            ),
            json=binding.jsonable_dict(
                by_alias=True, include={"binding_key": ..., "is_default_engine": ...}
            ),
        )
        return Binding.parse_model(response.json()["binding"])



from __future__ import annotations

import logging
import time
from datetime import datetime
from typing import TYPE_CHECKING, Any, Optional, Sequence

from pydantic import Field, PrivateAttr

from firebolt.model.V1 import FireboltBaseModel
from firebolt.model.V1.engine_revision import EngineRevisionKey
from firebolt.model.V1.region import RegionKey
from firebolt.service.V1.types import (
    EngineStatus,
    EngineStatusSummary,
    EngineType,
    WarmupMethod,
)
from firebolt.utils.urls import (
    ACCOUNT_ENGINE_START_URL,
    ACCOUNT_ENGINE_STOP_URL,
)

if TYPE_CHECKING:
    from firebolt.service.V1.engine import EngineService

logger = logging.getLogger(__name__)


class EngineKey(FireboltBaseModel):
    account_id: str
    engine_id: str


def wait(seconds: int, timeout_time: float, error_message: str, verbose: bool) -> None:
    time.sleep(seconds)
    if time.time() > timeout_time:
        raise TimeoutError(error_message)
    if verbose:
        print(".", end="")


class EngineSettings(FireboltBaseModel):
    """
    Engine settings.

    See also: :py:class:`EngineRevisionSpecification
    <firebolt.model.engine_revision.EngineRevisionSpecification>`
    which also contains engine configuration.
    """

    preset: str
    auto_stop_delay_duration: str = Field(pattern=r"^[0-9]+[sm]$|^0$")
    minimum_logging_level: str
    is_read_only: bool
    warm_up: str

    @classmethod
    def default(
        cls,
        engine_type: EngineType = EngineType.GENERAL_PURPOSE,
        auto_stop_delay_duration: str = "1200s",
        warm_up: WarmupMethod = WarmupMethod.PRELOAD_INDEXES,
        minimum_logging_level: str = "ENGINE_SETTINGS_LOGGING_LEVEL_INFO",
    ) -> EngineSettings:
        if engine_type == EngineType.GENERAL_PURPOSE:
            preset = engine_type.GENERAL_PURPOSE.api_settings_preset_name  # type: ignore # noqa: E501
            is_read_only = False
        else:
            preset = engine_type.DATA_ANALYTICS.api_settings_preset_name  # type: ignore
            is_read_only = True

        return cls(
            preset=preset,
            auto_stop_delay_duration=auto_stop_delay_duration,
            minimum_logging_level=minimum_logging_level,
            is_read_only=is_read_only,
            warm_up=warm_up.api_name,
        )


class FieldMask(FireboltBaseModel):
    paths: Sequence[str] = Field(alias="paths")


class Engine(FireboltBaseModel):
    """
    A Firebolt engine. Responsible for performing work (queries, ingestion).

    Engines are configured in :py:class:`Settings
    <firebolt.model.engine.EngineSettings>`
    and in :py:class:`EngineRevisionSpecification
    <firebolt.model.engine_revision.EngineRevisionSpecification>`.
    """

    # internal
    _service: EngineService = PrivateAttr()

    # required
    name: str = Field(min_length=1, max_length=255, pattern=r"^[0-9a-zA-Z_]+$")
    compute_region_key: RegionKey = Field(alias="compute_region_id")
    settings: EngineSettings

    # optional
    key: Optional[EngineKey] = Field(None, alias="id")
    description: Optional[str] = None
    emoji: Optional[str] = None
    current_status: Optional[EngineStatus] = None
    current_status_summary: Optional[EngineStatusSummary] = None
    latest_revision_key: Optional[EngineRevisionKey] = Field(
        None, alias="latest_revision_id"
    )
    endpoint: Optional[str] = None
    endpoint_serving_revision_key: Optional[EngineRevisionKey] = Field(
        None, alias="endpoint_serving_revision_id"
    )
    create_time: Optional[datetime] = None
    create_actor: Optional[str] = None
    last_update_time: Optional[datetime] = None
    last_update_actor: Optional[str] = None
    last_use_time: Optional[datetime] = None
    desired_status: Optional[str] = None
    health_status: Optional[str] = None
    endpoint_desired_revision_key: Optional[EngineRevisionKey] = Field(
        None, alias="endpoint_desired_revision_id"
    )

    @classmethod
    def parse_obj_with_service(cls, obj: Any, engine_service: EngineService) -> Engine:
        engine = cls.parse_model(obj)
        engine._service = engine_service
        return engine

    @property
    def engine_id(self) -> str:
        if self.key is None:
            raise ValueError("engine key is None")
        return self.key.engine_id

    def get_latest(self) -> Engine:
        """Get an up-to-date instance of the engine from Firebolt."""
        return self._service.get(id_=self.engine_id)

    def start(
        self,
        wait_for_startup: bool = True,
        wait_timeout_seconds: int = 3600,
        verbose: bool = False,
    ) -> Engine:
        """
        Start an engine. If it's already started, do nothing.

        Args:
            wait_for_startup:
                If True, wait for startup to complete.
                If False, return immediately after requesting startup.
            wait_timeout_seconds:
                Number of seconds to wait for startup to complete
                before raising a TimeoutError
            verbose:
                If True, print dots periodically while waiting for engine start.
                If False, do not print any dots.

        Returns:
            The updated engine from Firebolt.
        """
        timeout_time = time.time() + wait_timeout_seconds

        engine = self.get_latest()
        if (
            engine.current_status_summary
            == EngineStatusSummary.ENGINE_STATUS_SUMMARY_RUNNING
        ):
            logger.info(
                f"Engine (engine_id={self.engine_id}, name={self.name}) "
                "is already running."
            )
            return engine

        # wait for engine to stop first, if it's already stopping
        # FUTURE: revisit logging and consider consolidating this if & the while below.
        elif (
            engine.current_status_summary
            == EngineStatusSummary.ENGINE_STATUS_SUMMARY_STOPPING
        ):
            logger.info(
                f"Engine (engine_id={engine.engine_id}, name={engine.name}) "
                "is in currently stopping, waiting for it to stop first."
            )
            while (
                engine.current_status_summary
                != EngineStatusSummary.ENGINE_STATUS_SUMMARY_STOPPED
            ):
                wait(
                    seconds=5,
                    timeout_time=timeout_time,
                    error_message=(
                        "Engine "
                        f"(engine_id={engine.engine_id}, name={engine.name}) "
                        f"did not stop within {wait_timeout_seconds} seconds."
                    ),
                    verbose=True,
                )
                engine = engine.get_latest()

            logger.info(
                f"Engine (engine_id={engine.engine_id}, name={engine.name}) stopped."
            )

        engine = self._send_engine_request(ACCOUNT_ENGINE_START_URL)
        logger.info(
            f"Starting Engine (engine_id={engine.engine_id}, name={engine.name})"
        )

        # wait for engine to start
        while wait_for_startup and engine.current_status_summary not in {
            EngineStatusSummary.ENGINE_STATUS_SUMMARY_RUNNING,
            EngineStatusSummary.ENGINE_STATUS_SUMMARY_FAILED,
        }:
            wait(
                seconds=5,
                timeout_time=timeout_time,
                error_message=(  # noqa: E501
                    f"Could not start engine within {wait_timeout_seconds} seconds."
                ),
                verbose=verbose,
            )
            previous_status_summary = engine.current_status_summary
            engine = engine.get_latest()
            if engine.current_status_summary != previous_status_summary:
                logger.info(
                    "Engine status_summary="
                    f"{getattr(engine.current_status_summary, 'name')}"
                )

        return engine

    def stop(
        self, wait_for_stop: bool = False, wait_timeout_seconds: int = 3600
    ) -> Engine:
        """Stop an Engine running on Firebolt."""
        timeout_time = time.time() + wait_timeout_seconds

        engine = self._send_engine_request(ACCOUNT_ENGINE_STOP_URL)
        logger.info(f"Stopping Engine (engine_id={self.engine_id}, name={self.name})")

        while wait_for_stop and engine.current_status_summary not in {
            EngineStatusSummary.ENGINE_STATUS_SUMMARY_STOPPED,
            EngineStatusSummary.ENGINE_STATUS_SUMMARY_FAILED,
        }:
            wait(
                seconds=5,
                timeout_time=timeout_time,
                error_message=(  # noqa: E501
                    f"Could not stop engine within {wait_timeout_seconds} seconds."
                ),
                verbose=False,
            )

            engine = engine.get_latest()

        return engine

    def _send_engine_request(self, url: str) -> Engine:
        response = self._service.client.post(
            url=url.format(
                account_id=self._service.account_id, engine_id=self.engine_id
            )
        )
        return Engine.parse_obj_with_service(
            obj=response.json()["engine"], engine_service=self._service
        )



import logging
from typing import Optional

from httpx import Timeout

from firebolt.client import (
    DEFAULT_API_URL,
    Auth,
    ClientV1,
    ClientV2,
    log_request,
    log_response,
    raise_on_4xx_5xx,
)
from firebolt.common import Settings
from firebolt.db import connect
from firebolt.service.V1.binding import BindingService
from firebolt.service.V1.database import DatabaseService as DatabaseServiceV1
from firebolt.service.V1.engine import EngineService as EngineServiceV1
from firebolt.service.V1.provider import get_provider_id
from firebolt.service.V1.region import RegionService
from firebolt.service.V2.database import DatabaseService as DatabaseServiceV2
from firebolt.service.V2.engine import EngineService as EngineServiceV2
from firebolt.service.V2.instance_type import InstanceTypeService
from firebolt.utils.util import fix_url_schema

DEFAULT_TIMEOUT_SECONDS: int = 60 * 2

logger = logging.getLogger(__name__)

SETTINGS_DEPRECATION_MESSAGE = """
Using Settings objects for ResourceManager intialization is deprecated.
Please pass parameters directly
Example:
 >>> rm = ResourceManager(auth=ClientCredentials(..), account_name="my_account", ..)
"""


class ResourceManager:
    """
    ResourceManager to access various Firebolt resources:

    - databases
    - engines

    Also provides listings of:

    - instance types (AWS instance types which engines can use)
    """

    __slots__ = (
        "account_name",
        "account_id",
        "api_endpoint",
        "_client",
        "_connection",
        "regions",
        "instance_types",
        "provider_id",
        "databases",
        "engines",
        "engine_revisions",
        "bindings",
        "default_region",
        "_version",
    )

    def __init__(
        self,
        settings: Optional[Settings] = None,
        auth: Optional[Auth] = None,
        account_name: Optional[str] = None,
        api_endpoint: str = DEFAULT_API_URL,
        # Legacy parameters
        default_region: Optional[str] = None,
    ):
        if settings:
            logger.warning(SETTINGS_DEPRECATION_MESSAGE)
            if auth or account_name or (api_endpoint != DEFAULT_API_URL):
                raise ValueError(
                    "Other ResourceManager parameters are not allowed "
                    "when Settings are provided"
                )
            auth = settings.auth
            account_name = settings.account_name
            api_endpoint = settings.server
            default_region = settings.default_region

        for param, name in ((auth, "auth"),):
            if not param:
                raise ValueError(f"Missing {name} value")

        # type checks
        assert auth is not None

        version = auth.get_firebolt_version()
        if version == 2:
            client_class = ClientV2
            assert account_name is not None
        elif version == 1:
            client_class = ClientV1
        else:
            raise ValueError(f"Unsupported Firebolt version: {version}")
        # Separate client for API calls, not via engine URL
        self._client = client_class(
            auth=auth,
            base_url=fix_url_schema(api_endpoint),
            account_name=account_name,  # type: ignore # already checked above for v2
            api_endpoint=api_endpoint,
            timeout=Timeout(DEFAULT_TIMEOUT_SECONDS),
            event_hooks={
                "request": [log_request],
                "response": [raise_on_4xx_5xx, log_response],
            },
        )
        # V1 does not use a DB connection
        if version != 1:
            self._connection = connect(
                auth=auth,
                account_name=account_name,
                api_endpoint=api_endpoint,
            )
        else:
            self._connection = None  # type: ignore
        self.account_name = account_name
        self.api_endpoint = api_endpoint
        self.account_id = self._client.account_id
        self.default_region = default_region
        self.provider_id: Optional[str] = None
        if version == 2:
            self._init_services_v2()
        elif version == 1:
            self._init_services_v1()

    def _init_services_v2(self) -> None:
        # avoid circular import

        # Cloud Platform Resources (AWS)
        self.instance_types = InstanceTypeService(resource_manager=self)

        # Firebolt Resources
        self.databases = DatabaseServiceV2(resource_manager=self)
        self.engines = EngineServiceV2(resource_manager=self)

        # Not applicable to V2
        self.provider_id = None

    def _init_services_v1(self) -> None:
        # Cloud Platform Resources (AWS)
        self.regions = RegionService(resource_manager=self)  # type: ignore

        # Firebolt Resources
        self.bindings = BindingService(resource_manager=self)  # type: ignore
        self.engines = EngineServiceV1(resource_manager=self)  # type: ignore
        self.databases = DatabaseServiceV1(resource_manager=self)  # type: ignore

        self.provider_id = get_provider_id(client=self._client)

    def __del__(self) -> None:
        if hasattr(self, "_client"):
            self._client.close()
        if hasattr(self, "_connection") and self._connection is not None:
            self._connection.close()



KEEPALIVE_FLAG: int = 1
KEEPIDLE_RATE: int = 60  # seconds
DEFAULT_TIMEOUT_SECONDS: int = 60

# Running statuses in infromation schema
ENGINE_STATUS_RUNNING_LIST = ["RUNNING", "Running", "ENGINE_STATE_RUNNING"]



from firebolt.client.auth.base import Auth
from firebolt.client.auth.client_credentials import ClientCredentials
from firebolt.client.auth.service_account import ServiceAccount
from firebolt.client.auth.token import Token
from firebolt.client.auth.username_password import UsernamePassword



from datetime import datetime
from typing import Optional

from pydantic import Field

from firebolt.model.V1 import FireboltBaseModel


class RegionKey(FireboltBaseModel, frozen=True):  # type: ignore
    provider_id: str
    region_id: str


class Region(FireboltBaseModel):
    key: RegionKey = Field(alias="id")
    name: str

    # optional
    display_name: Optional[str] = None
    create_time: Optional[datetime] = None
    last_update_time: Optional[datetime] = None



from __future__ import annotations

import logging
from dataclasses import dataclass, field
from datetime import datetime
from typing import TYPE_CHECKING, ClassVar, List

from firebolt.model.V2 import FireboltBaseModel
from firebolt.service.V2.types import EngineStatus
from firebolt.utils.exception import AttachedEngineInUseError

if TYPE_CHECKING:
    from firebolt.model.V2.engine import Engine
    from firebolt.service.V2.database import DatabaseService

logger = logging.getLogger(__name__)


@dataclass
class Database(FireboltBaseModel):
    """
    A Firebolt database.

    Databases belong to a region and have a description,
    but otherwise are not configurable.
    """

    ALTER_SQL: ClassVar[str] = 'ALTER DATABASE "{}" SET DESCRIPTION = ?'

    DROP_SQL: ClassVar[str] = 'DROP DATABASE "{}"'

    # internal
    _service: DatabaseService = field(repr=False, compare=False)

    # required
    name: str = field(metadata={"db_name": "catalog_name"})
    description: str = field()
    create_time: datetime = field(metadata={"db_name": "created"})
    create_actor: str = field(metadata={"db_name": "catalog_owner"})

    def get_attached_engines(self) -> List[Engine]:
        """Get a list of engines that are attached to this database."""
        return self._service.resource_manager.engines.get_many(database_name=self.name)

    def attach_engine(self, engine: Engine) -> None:
        """
        Attach an engine to this database.

        Args:
            engine: The engine to attach.
        """
        return self._service.resource_manager.engines.attach_to_database(
            engine.name, self.name
        )

    def update(self, description: str) -> Database:
        """
        Updates a database description.
        """
        if not description:
            return self

        for engine in self.get_attached_engines():
            if engine.current_status not in {
                EngineStatus.RUNNING,
                EngineStatus.STOPPED,
            }:
                raise AttachedEngineInUseError(method_name="update")

        sql = self.ALTER_SQL.format(self.name)
        with self._service._connection.cursor() as c:
            c.execute(sql, (description,))
        self.description = description
        return self

    def delete(self) -> None:
        """
        Delete a database from Firebolt.

        Raises an error if there are any attached engines.
        """

        for engine in self.get_attached_engines():
            if engine.current_status in {
                EngineStatus.STARTING,
                EngineStatus.STOPPING,
            }:
                raise AttachedEngineInUseError(method_name="delete")

        with self._service._connection.cursor() as c:
            c.execute(self.DROP_SQL.format(self.name))



from __future__ import annotations

import logging
import time
from abc import ABCMeta, abstractmethod
from functools import wraps
from types import TracebackType
from typing import (
    TYPE_CHECKING,
    Any,
    Dict,
    Iterator,
    List,
    Optional,
    Sequence,
    Union,
)
from urllib.parse import urljoin

from httpx import (
    URL,
    USE_CLIENT_DEFAULT,
    Headers,
    Response,
    TimeoutException,
    codes,
)

from firebolt.client.client import AsyncClient, AsyncClientV1, AsyncClientV2
from firebolt.common._types import (
    ColType,
    ParameterType,
    SetParameter,
    split_format_sql,
)
from firebolt.common.base_cursor import (
    JSON_OUTPUT_FORMAT,
    RESET_SESSION_HEADER,
    UPDATE_ENDPOINT_HEADER,
    UPDATE_PARAMETERS_HEADER,
    BaseCursor,
    CursorState,
    _parse_update_endpoint,
    _parse_update_parameters,
    _raise_if_internal_set_parameter,
    async_not_allowed,
    check_not_closed,
    check_query_executed,
)
from firebolt.utils.exception import (
    EngineNotRunningError,
    FireboltDatabaseError,
    FireboltError,
    NotSupportedError,
    OperationalError,
    ProgrammingError,
    QueryTimeoutError,
)
from firebolt.utils.timeout_controller import TimeoutController
from firebolt.utils.urls import DATABASES_URL, ENGINES_URL

if TYPE_CHECKING:
    from firebolt.async_db.connection import Connection

from firebolt.utils.util import _print_error_body, raise_errors_from_body

logger = logging.getLogger(__name__)


class Cursor(BaseCursor, metaclass=ABCMeta):
    """
    Class, responsible for executing queries to Firebolt Database.
    Should not be created directly,
    use :py:func:`connection.cursor <firebolt.async_db.connection.Connection>`

    Args:
        description: Information about a single result row
        rowcount: The number of rows produced by last query
        closed: True if connection is closed, False otherwise
        arraysize: Read/Write, specifies the number of rows to fetch at a time
            with the :py:func:`fetchmany` method
    """

    def __init__(
        self,
        *args: Any,
        client: AsyncClient,
        connection: Connection,
        **kwargs: Any,
    ) -> None:
        super().__init__(*args, **kwargs)
        self._client = client
        self.connection = connection
        self.engine_url = connection.engine_url
        if connection.init_parameters:
            self._update_set_parameters(connection.init_parameters)

    async def _api_request(
        self,
        query: str = "",
        parameters: Optional[dict[str, Any]] = None,
        path: str = "",
        use_set_parameters: bool = True,
        timeout: Optional[float] = None,
    ) -> Response:
        """
        Query API, return Response object.

        Args:
            query (str): SQL query
            parameters (Optional[Sequence[ParameterType]]): A sequence of substitution
                parameters. Used to replace '?' placeholders inside a query with
                actual values. Note: In order to "output_format" dict value, it
                    must be an empty string. If no value not specified,
                    JSON_OUTPUT_FORMAT will be used.
            path (str): endpoint suffix, for example "cancel" or "status"
            use_set_parameters: Optional[bool]: Some queries will fail if additional
                set parameters are sent. Setting this to False will allow
                self._set_parameters to be ignored.
            timeout (Optional[float]): Request execution timeout in seconds
        """
        parameters = parameters or {}
        if use_set_parameters:
            parameters = {**(self._set_parameters or {}), **parameters}
        if self.parameters:
            parameters = {**self.parameters, **parameters}
        try:
            return await self._client.request(
                url=urljoin(self.engine_url.rstrip("/") + "/", path or ""),
                method="POST",
                params=parameters,
                content=query,
                timeout=timeout if timeout is not None else USE_CLIENT_DEFAULT,
            )
        except TimeoutException:
            raise QueryTimeoutError()

    async def _raise_if_error(self, resp: Response) -> None:
        """Raise a proper error if any"""
        if resp.status_code == codes.INTERNAL_SERVER_ERROR:
            raise OperationalError(
                f"Error executing query:\n{resp.read().decode('utf-8')}"
            )
        if resp.status_code == codes.FORBIDDEN:
            if self.database and not await self.is_db_available(self.database):
                raise FireboltDatabaseError(f"Database {self.database} does not exist")
            raise ProgrammingError(resp.read().decode("utf-8"))
        if (
            resp.status_code == codes.SERVICE_UNAVAILABLE
            or resp.status_code == codes.NOT_FOUND
        ) and not await self.is_engine_running(self.engine_url):
            raise EngineNotRunningError(
                f"Firebolt engine {self.engine_url} "
                "needs to be running to run queries against it."
            )
        raise_errors_from_body(resp)
        # If no structure for error is found, log the body and raise the error
        _print_error_body(resp)
        resp.raise_for_status()

    async def _validate_set_parameter(
        self, parameter: SetParameter, timeout: Optional[float]
    ) -> None:
        """Validate parameter by executing simple query with it."""
        _raise_if_internal_set_parameter(parameter)
        resp = await self._api_request(
            "select 1", {parameter.name: parameter.value}, timeout=timeout
        )
        # Handle invalid set parameter
        if resp.status_code == codes.BAD_REQUEST:
            raise OperationalError(resp.text)
        await self._raise_if_error(resp)

        # set parameter passed validation
        self._set_parameters[parameter.name] = parameter.value

    async def _parse_response_headers(self, headers: Headers) -> None:
        if headers.get(UPDATE_ENDPOINT_HEADER):
            endpoint, params = _parse_update_endpoint(
                headers.get(UPDATE_ENDPOINT_HEADER)
            )
            self._update_set_parameters(params)
            self.engine_url = endpoint

        if headers.get(RESET_SESSION_HEADER):
            self.flush_parameters()

        if headers.get(UPDATE_PARAMETERS_HEADER):
            param_dict = _parse_update_parameters(headers.get(UPDATE_PARAMETERS_HEADER))
            self._update_set_parameters(param_dict)

    @abstractmethod
    async def execute_async(
        self,
        query: str,
        parameters: Optional[Sequence[ParameterType]] = None,
        skip_parsing: bool = False,
    ) -> int:
        """Execute a database query without maintaining a connection."""
        ...

    async def _do_execute(
        self,
        raw_query: str,
        parameters: Sequence[Sequence[ParameterType]],
        skip_parsing: bool = False,
        timeout: Optional[float] = None,
        async_execution: bool = False,
    ) -> None:
        self._reset()
        queries: List[Union[SetParameter, str]] = (
            [raw_query] if skip_parsing else split_format_sql(raw_query, parameters)
        )
        timeout_controller = TimeoutController(timeout)

        if len(queries) > 1 and async_execution:
            raise FireboltError(
                "Server side async does not support multi-statement queries"
            )
        try:
            for query in queries:
                await self._execute_single_query(
                    query, timeout_controller, async_execution
                )
            self._state = CursorState.DONE
        except Exception:
            self._state = CursorState.ERROR
            raise

    async def _execute_single_query(
        self,
        query: Union[SetParameter, str],
        timeout_controller: TimeoutController,
        async_execution: bool,
    ) -> None:
        start_time = time.time()
        Cursor._log_query(query)
        timeout_controller.raise_if_timeout()

        if isinstance(query, SetParameter):
            if async_execution:
                raise FireboltError(
                    "Server side async does not support set statements, "
                    "please use execute to set this parameter"
                )
            await self._validate_set_parameter(query, timeout_controller.remaining())
        else:
            await self._handle_query_execution(
                query, timeout_controller, async_execution
            )

        if not async_execution:
            logger.info(
                f"Query fetched {self.rowcount} rows in"
                f" {time.time() - start_time} seconds."
            )
        else:
            logger.info("Query submitted for async execution.")

    async def _handle_query_execution(
        self, query: str, timeout_controller: TimeoutController, async_execution: bool
    ) -> None:
        query_params: Dict[str, Any] = {"output_format": JSON_OUTPUT_FORMAT}
        if async_execution:
            query_params["async"] = True
        resp = await self._api_request(
            query,
            query_params,
            timeout=timeout_controller.remaining(),
        )
        await self._raise_if_error(resp)
        if async_execution:
            self._parse_async_response(resp)
        else:
            await self._parse_response_headers(resp.headers)
            row_set = self._row_set_from_response(resp)
            self._append_row_set(row_set)

    @check_not_closed
    async def execute(
        self,
        query: str,
        parameters: Optional[Sequence[ParameterType]] = None,
        skip_parsing: bool = False,
        timeout_seconds: Optional[float] = None,
    ) -> Union[int, str]:
        """Prepare and execute a database query.

        Supported features:
            Parameterized queries: placeholder characters ('?') are substituted
                with values provided in `parameters`. Values are formatted to
                be properly recognized by database and to exclude SQL injection.
            Multi-statement queries: multiple statements, provided in a single query
                and separated by semicolon, are executed separatelly and sequentially.
                To switch to next statement result, `nextset` method should be used.
            SET statements: to provide additional query execution parameters, execute
                `SET param=value` statement before it. All parameters are stored in
                cursor object until it's closed. They can also be removed with
                `flush_parameters` method call.

        Args:
            query (str): SQL query to execute
            parameters (Optional[Sequence[ParameterType]]): A sequence of substitution
                parameters. Used to replace '?' placeholders inside a query with
                actual values
            skip_parsing (bool): Flag to disable query parsing. This will
                disable parameterized, multi-statement and SET queries,
                while improving performance
            timeout_seconds (Optional[float]): Query execution timeout in seconds

        Returns:
            int: Query row count.
        """
        params_list = [parameters] if parameters else []
        await self._do_execute(
            query, params_list, skip_parsing, timeout=timeout_seconds
        )
        return self.rowcount

    @check_not_closed
    async def executemany(
        self,
        query: str,
        parameters_seq: Sequence[Sequence[ParameterType]],
        timeout_seconds: Optional[float] = None,
    ) -> Union[int, str]:
        """Prepare and execute a database query.

        Supports providing multiple substitution parameter sets, executing them
        as multiple statements sequentially.

        Supported features:
            Parameterized queries: Placeholder characters ('?') are substituted
                with values provided in `parameters`. Values are formatted to
                be properly recognized by database and to exclude SQL injection.
            Multi-statement queries: Multiple statements, provided in a single query
                and separated by semicolon, are executed separately and sequentially.
                To switch to next statement result, use `nextset` method.
            SET statements: To provide additional query execution parameters, execute
                `SET param=value` statement before it. All parameters are stored in
                cursor object until it's closed. They can also be removed with
                `flush_parameters` method call.

        Args:
            query (str): SQL query to execute.
            parameters_seq (Sequence[Sequence[ParameterType]]): A sequence of
               substitution parameter sets. Used to replace '?' placeholders inside a
               query with actual values from each set in a sequence. Resulting queries
               for each subset are executed sequentially.
            timeout_seconds (Optional[float]): Query execution timeout in seconds.

        Returns:
            int: Query row count.
        """
        await self._do_execute(query, parameters_seq, timeout=timeout_seconds)
        return self.rowcount

    @abstractmethod
    async def is_db_available(self, database: str) -> bool:
        """Verify that the database exists."""
        ...

    @abstractmethod
    async def is_engine_running(self, engine_url: str) -> bool:
        """Verify that the engine is running."""
        ...

    @wraps(BaseCursor.fetchone)
    async def fetchone(self) -> Optional[List[ColType]]:
        """Fetch the next row of a query result set."""
        return super().fetchone()

    @wraps(BaseCursor.fetchmany)
    async def fetchmany(self, size: Optional[int] = None) -> List[List[ColType]]:
        """
        Fetch the next set of rows of a query result;
        size is cursor.arraysize by default.
        """
        return super().fetchmany(size)

    @wraps(BaseCursor.fetchall)
    async def fetchall(self) -> List[List[ColType]]:
        """Fetch all remaining rows of a query result."""
        return super().fetchall()

    @wraps(BaseCursor.nextset)
    async def nextset(self) -> None:
        return super().nextset()

    # Iteration support
    @check_not_closed
    @async_not_allowed
    @check_query_executed
    def __aiter__(self) -> Cursor:
        return self

    # TODO: figure out how to implement __aenter__ and __await__
    @check_not_closed
    def __aenter__(self) -> Cursor:
        return self

    @check_not_closed
    def __enter__(self) -> Cursor:
        return self

    def __exit__(
        self, exc_type: type, exc_val: Exception, exc_tb: TracebackType
    ) -> None:
        self.close()

    def __await__(self) -> Iterator:
        yield None

    async def __aexit__(
        self, exc_type: type, exc_val: Exception, exc_tb: TracebackType
    ) -> None:
        self.close()

    @check_not_closed
    @async_not_allowed
    @check_query_executed
    async def __anext__(self) -> List[ColType]:
        row = await self.fetchone()
        if row is None:
            raise StopAsyncIteration
        return row


class CursorV2(Cursor):
    def __init__(
        self,
        *args: Any,
        client: AsyncClientV2,
        connection: Connection,
        **kwargs: Any,
    ) -> None:
        assert isinstance(client, AsyncClientV2)
        super().__init__(*args, client=client, connection=connection, **kwargs)

    @check_not_closed
    async def execute_async(
        self,
        query: str,
        parameters: Optional[Sequence[ParameterType]] = None,
        skip_parsing: bool = False,
    ) -> int:
        """
        Execute a database query without maintating a connection.

        Supported features:
            Parameterized queries: placeholder characters ('?') are substituted
                with values provided in `parameters`. Values are formatted to
                be properly recognized by database and to exclude SQL injection.

        Not supported:
            Multi-statement queries: multiple statements, provided in a single query
                and separated by semicolon.
            SET statements: to provide additional query execution parameters, execute
                `SET param=value` statement before it. Use `execute` method to set
                parameters.

        Args:
            query (str): SQL query to execute
            parameters (Optional[Sequence[ParameterType]]): A sequence of substitution
                parameters. Used to replace '?' placeholders inside a query with
                actual values
            skip_parsing (bool): Flag to disable query parsing. This will
                disable parameterized queries while potentially improving performance

        Returns:
            int: Always returns -1, as async execution does not return row count.
        """
        await self._do_execute(
            query,
            [parameters] if parameters else [],
            skip_parsing,
            async_execution=True,
        )
        return -1

    async def is_db_available(self, database_name: str) -> bool:
        """
        Verify that the database exists.

        Args:
            connection (firebolt.async_db.connection.Connection)
            database_name (str): Name of a database
        """
        # For v2 accounts if we're connected it automatically
        # means the database is available
        return True

    async def is_engine_running(self, engine_url: str) -> bool:
        """
        Verify that the engine is running.

        Args:
            connection (firebolt.async_db.connection.Connection): connection.
            engine_url (str): URL of the engine
        """
        # For v2 accounts we don't have the engine context,
        # so we can't check if it's running
        return True


class CursorV1(Cursor):
    def __init__(
        self,
        *args: Any,
        client: AsyncClientV1,
        connection: Connection,
        **kwargs: Any,
    ) -> None:
        assert isinstance(client, AsyncClientV1)
        super().__init__(*args, client=client, connection=connection, **kwargs)

    async def is_db_available(self, database_name: str) -> bool:
        """
        Verify that the database exists.

        Args:
            connection (firebolt.async_db.connection.Connection)
        """
        resp = await self._filter_request(
            DATABASES_URL, {"filter.name_contains": database_name}
        )
        return len(resp.json()["edges"]) > 0

    async def is_engine_running(self, engine_url: str) -> bool:
        """
        Verify that the engine is running.

        Args:
            connection (firebolt.async_db.connection.Connection): connection.
        """
        # Url is not guaranteed to be of this structure,
        # but for the sake of error checking this is sufficient.
        engine_name = URL(engine_url).host.split(".")[0].replace("-", "_")
        resp = await self._filter_request(
            ENGINES_URL,
            {
                "filter.name_contains": engine_name,
                "filter.current_status_eq": "ENGINE_STATUS_RUNNING_REVISION_SERVING",
            },
        )
        return len(resp.json()["edges"]) > 0

    async def _filter_request(self, endpoint: str, filters: dict) -> Response:
        resp = await self.connection._client.request(
            # Full url overrides the client url, which contains engine as a prefix.
            url=self.connection.api_endpoint + endpoint,
            method="GET",
            params=filters,
        )
        resp.raise_for_status()
        return resp

    async def execute_async(
        self,
        query: str,
        parameters: Optional[Sequence[ParameterType]] = None,
        skip_parsing: bool = False,
    ) -> int:
        raise NotSupportedError(
            "Async execution is not supported in this version " " of Firebolt."
        )



from logging import getLogger
from typing import List, Optional, Tuple, Union

from firebolt.model.V2.engine import Database, Engine
from firebolt.model.V2.instance_type import InstanceType
from firebolt.service.V2.base import BaseService
from firebolt.service.V2.types import EngineStatus
from firebolt.utils.exception import EngineNotFoundError

logger = getLogger(__name__)


class EngineService(BaseService):
    DB_FIELDS = (
        "engine_name",
        "region",
        "type",
        "nodes",
        "status",
        "default_database",
        "version",
        "url",
        "warmup",
        "auto_stop",
        "engine_type",
    )

    GET_SQL = "SELECT {} FROM information_schema.engines"
    GET_BY_NAME_SQL = GET_SQL + " WHERE engine_name=?"
    GET_WHERE_SQL = " WHERE "

    CREATE_PREFIX_SQL = 'CREATE ENGINE {}"{}"'
    IF_NOT_EXISTS_SQL = "IF NOT EXISTS "
    CREATE_WITH_SQL = " WITH "

    CREATE_PARAMETER_NAMES = (
        "",
        "",
        "TYPE",
        "NODES",
        "AUTO_STOP",
        "",
    )

    ATTACH_TO_DB_SQL = 'ALTER ENGINE "{}" SET DEFAULT_DATABASE="{}"'
    DISALLOWED_ACCOUNT_V2_PARAMETERS = [
        "region",
        "engine_type",
        "warmup",
    ]

    def _get_dict(self, name: str) -> dict:
        with self._connection.cursor() as c:
            sql = self.GET_BY_NAME_SQL.format(", ".join(self.DB_FIELDS))
            count = c.execute(sql, (name,))
            if count == 0:
                raise EngineNotFoundError(name)
            return {
                column.name: value for column, value in zip(c.description, c.fetchone())
            }

    def get(self, name: str) -> Engine:
        """Get an engine from Firebolt by its name."""
        return Engine._from_dict(self._get_dict(name), self)

    def get_by_name(self, name: str) -> Engine:
        return self.get(name)

    def get_many(
        self,
        name_contains: Optional[str] = None,
        current_status_eq: Union[str, EngineStatus, None] = None,
        current_status_not_eq: Union[str, EngineStatus, None] = None,
        region_eq: Optional[str] = None,
        database_name: Optional[str] = None,
    ) -> List[Engine]:
        """
        Get a list of engines on Firebolt.

        Args:
            name_contains: Filter for engines with a name containing this substring
            current_status_eq: Filter for engines with this status
            current_status_not_eq: Filter for engines that do not have this status
            region_eq: Filter for engines by region

        Returns:
            A list of engines matching the filters
        """

        if region_eq:
            raise ValueError(
                "Parameter 'region_eq' is not supported in this version of Firebolt."
            )

        sql = self.GET_SQL.format(", ".join(self.DB_FIELDS))
        parameters = []
        if any(
            (
                name_contains,
                current_status_eq,
                current_status_not_eq,
                database_name,
            )
        ):
            condition = []
            if name_contains:
                condition.append("engine_name like ?")
                parameters.append(f"%{name_contains}%")
            if current_status_eq:
                condition.append("status = ?")
                parameters.append(str(current_status_eq))
            if current_status_not_eq:
                condition.append("status != ?")
                parameters.append(str(current_status_eq))
            if database_name:
                condition.append("default_database = ?")
                parameters.append(database_name)
            sql += self.GET_WHERE_SQL + " AND ".join(condition)

        with self._connection.cursor() as c:
            c.execute(sql, parameters)
            dicts = [
                {column.name: value for column, value in zip(c.description, row)}
                for row in c.fetchall()
            ]
            return [Engine._from_dict(_dict, self) for _dict in dicts]

    @staticmethod
    def _format_engine_parameter(
        value: Union[str, int, InstanceType]
    ) -> Union[str, int]:
        if not isinstance(value, (str, int, InstanceType)) or isinstance(value, bool):
            raise TypeError(f"Unsupported type {type(value)} for engine parameter. ")
        if isinstance(value, InstanceType):
            return value.name
        return value

    def _format_engine_attribute_sql(
        self, param: str, value: Union[str, int, InstanceType]
    ) -> Tuple[str, List]:
        """Format an engine attribute for use in a SQL query.

        Args:
            param: The name of the parameter
            value: The value of the parameter

        Returns:
            A tuple containing the formatted SQL string and a list of parameters
            to pass for execution
        """
        if param == "TYPE":
            return f"{param} = {self._format_engine_parameter(value)} ", []
        return f"{param} = ? ", [self._format_engine_parameter(value)]

    def create(
        self,
        name: str,
        region: Optional[str] = None,
        engine_type: Optional[str] = None,
        spec: Union[InstanceType, str, None] = None,
        scale: Optional[int] = None,
        auto_stop: Optional[int] = None,
        warmup: Optional[str] = None,
        fail_if_exists: bool = True,
    ) -> Engine:
        """
        Create a new engine.

        Args:
            name: An identifier that specifies the name of the engine
            region: The AWS region in which the engine runs
            engine_type: The engine type. GENERAL_PURPOSE or DATA_ANALYTICS
            spec: Firebolt instance type. If not set, will default to
                the cheapest instance.
            scale: The number of compute instances on the engine.
                The scale can be any int from 1 to 128.
            auto_stop: The amount of time (in minutes)
            after which the engine automatically stops
            warmup: The warmup method that should be used:

                `MINIMAL` - On-demand loading (both indexes and tables' data)

                `PRELOAD_INDEXES` - Load indexes only

                `PRELOAD_ALL_DATA` - Full data auto-load
                (both indexes and table data - full warmup)
            fail_if_exists: Fail is an engine with provided name already exists

        Returns:
            Engine with the specified settings
        """

        logger.info(f"Creating engine {name}")

        disallowed_parameters = [
            name
            for name, value in (
                ("region", region),
                ("engine_type", engine_type),
                ("warmup", warmup),
            )
            if value
        ]
        if disallowed_parameters:
            raise ValueError(
                f"Parameters {disallowed_parameters} are not supported for this account"
            )

        sql = self.CREATE_PREFIX_SQL.format(
            ("" if fail_if_exists else self.IF_NOT_EXISTS_SQL), name
        )
        parameters = []
        if any(x is not None for x in (spec, scale, auto_stop)):
            sql += self.CREATE_WITH_SQL
            for param, value in zip(
                self.CREATE_PARAMETER_NAMES,
                (region, engine_type, spec, scale, auto_stop, warmup),
            ):
                if value is not None:
                    sql_part, new_params = self._format_engine_attribute_sql(
                        param, value
                    )
                    sql += sql_part
                    parameters.extend(new_params)
        with self._connection.cursor() as c:
            c.execute(sql, parameters)
        return self.get(name)

    def attach_to_database(
        self, engine: Union[Engine, str], database: Union[Database, str]
    ) -> None:
        engine_name = engine.name if isinstance(engine, Engine) else engine
        database_name = database.name if isinstance(database, Database) else database
        with self._connection.cursor() as c:
            c.execute(self.ATTACH_TO_DB_SQL.format(engine_name, database_name))
        if isinstance(engine, Engine):
            engine._database_name = (
                database.name if isinstance(database, Database) else database
            )



from __future__ import annotations

import logging
from types import TracebackType
from typing import Any, Dict, List, Optional, Type
from warnings import warn

from httpx import Timeout

from firebolt.client import DEFAULT_API_URL, Client, ClientV1, ClientV2
from firebolt.client.auth import Auth
from firebolt.common.base_connection import (
    ASYNC_QUERY_STATUS_REQUEST,
    ASYNC_QUERY_STATUS_RUNNING,
    ASYNC_QUERY_STATUS_SUCCESSFUL,
    BaseConnection,
)
from firebolt.common.cache import _firebolt_system_engine_cache
from firebolt.common.constants import DEFAULT_TIMEOUT_SECONDS
from firebolt.db.cursor import Cursor, CursorV1, CursorV2
from firebolt.db.util import _get_system_engine_url_and_params
from firebolt.utils.exception import (
    ConfigurationError,
    ConnectionClosedError,
    FireboltError,
)
from firebolt.utils.usage_tracker import get_user_agent_header
from firebolt.utils.util import fix_url_schema, validate_engine_name_and_url_v1

logger = logging.getLogger(__name__)


def connect(
    auth: Optional[Auth] = None,
    account_name: Optional[str] = None,
    database: Optional[str] = None,
    engine_name: Optional[str] = None,
    engine_url: Optional[str] = None,
    api_endpoint: str = DEFAULT_API_URL,
    disable_cache: bool = False,
    additional_parameters: Dict[str, Any] = {},
) -> Connection:
    # auth parameter is optional in function signature
    # but is required to connect.
    # PEP 249 recommends making it kwargs.
    if not auth:
        raise ConfigurationError("auth is required to connect.")

    # Type checks
    assert auth is not None
    user_drivers = additional_parameters.get("user_drivers", [])
    user_clients = additional_parameters.get("user_clients", [])
    user_agent_header = get_user_agent_header(user_drivers, user_clients)
    auth_version = auth.get_firebolt_version()
    if disable_cache:
        _firebolt_system_engine_cache.disable()
    # Use v2 if auth is ClientCredentials
    # Use v1 if auth is ServiceAccount or UsernamePassword
    if auth_version == 2:
        assert account_name is not None
        return connect_v2(
            auth=auth,
            user_agent_header=user_agent_header,
            account_name=account_name,
            database=database,
            engine_name=engine_name,
            api_endpoint=api_endpoint,
        )
    elif auth_version == 1:
        return connect_v1(
            auth=auth,
            user_agent_header=user_agent_header,
            account_name=account_name,
            database=database,
            engine_name=engine_name,
            engine_url=engine_url,
            api_endpoint=api_endpoint,
        )
    else:
        raise ConfigurationError(f"Unsupported auth type: {type(auth)}")


def connect_v2(
    auth: Auth,
    user_agent_header: str,
    account_name: Optional[str] = None,
    database: Optional[str] = None,
    engine_name: Optional[str] = None,
    api_endpoint: str = DEFAULT_API_URL,
) -> Connection:
    """Connect to Firebolt.

    Args:
        `auth` (Auth) Authentication object
        `database` (str): Name of the database to connect
        `engine_name` (Optional[str]): Name of the engine to connect to
        `account_name` (Optional[str]): For customers with multiple accounts;
                                        if none, default is used
        `api_endpoint` (str): Firebolt API endpoint. Used for authentication
        `additional_parameters` (Optional[Dict]): Dictionary of less widely-used
                                arguments for connection

    """
    # These parameters are optional in function signature
    # but are required to connect.
    # PEP 249 recommends making them kwargs.
    for name, value in [("account_name", account_name)]:
        if not value:
            raise ConfigurationError(f"{name} is required to connect.")

    # Type checks
    assert auth is not None
    assert account_name is not None

    api_endpoint = fix_url_schema(api_endpoint)

    system_engine_url, system_engine_params = _get_system_engine_url_and_params(
        auth, account_name, api_endpoint
    )

    client = ClientV2(
        auth=auth,
        account_name=account_name,
        api_endpoint=api_endpoint,
        timeout=Timeout(DEFAULT_TIMEOUT_SECONDS, read=None),
        headers={"User-Agent": user_agent_header},
    )

    with Connection(
        system_engine_url,
        None,
        client,
        CursorV2,
        api_endpoint,
        system_engine_params,
    ) as system_engine_connection:

        cursor = system_engine_connection.cursor()
        if database:
            cursor.execute(f'USE DATABASE "{database}"')
        if engine_name:
            cursor.execute(f'USE ENGINE "{engine_name}"')
        # Ensure cursors created from this connection are using the same starting
        # database and engine
        return Connection(
            cursor.engine_url,
            cursor.database,
            client.clone(),
            CursorV2,
            api_endpoint,
            cursor.parameters,
        )


class Connection(BaseConnection):
    """
    Firebolt database connection class. Implements PEP-249.

    Args:

        engine_url: Firebolt database engine REST API url
        database: Firebolt database name
        username: Firebolt account username
        password: Firebolt account password
        api_endpoint: Optional. Firebolt API endpoint. Used for authentication.

    Note:
        Firebolt currenly doesn't support transactions so commit and rollback methods
        are not implemented.
    """

    client_class: type
    cursor_type: Type[Cursor]
    __slots__ = (
        "_client",
        "_cursors",
        "engine_url",
        "api_endpoint",
        "_is_closed",
        "client_class",
        "cursor_type",
    )

    def __init__(
        self,
        engine_url: str,
        database: Optional[str],
        client: Client,
        cursor_type: Type[Cursor],
        api_endpoint: str = DEFAULT_API_URL,
        init_parameters: Optional[Dict[str, Any]] = None,
    ):
        super().__init__(cursor_type)
        self.api_endpoint = api_endpoint
        self.engine_url = engine_url
        self._cursors: List[Cursor] = []
        self._client = client
        self.init_parameters = init_parameters or {}
        if database:
            self.init_parameters["database"] = database

    def cursor(self, **kwargs: Any) -> Cursor:
        if self.closed:
            raise ConnectionClosedError("Unable to create cursor: connection closed.")

        c = self.cursor_type(client=self._client, connection=self, **kwargs)
        self._cursors.append(c)
        return c

    def _remove_cursor(self, cursor: Cursor) -> None:
        # This way it's atomic
        try:
            self._cursors.remove(cursor)
        except ValueError:
            pass

    def close(self) -> None:
        if self.closed:
            return

        cursors = self._cursors[:]
        for c in cursors:
            # Here c can already be closed by another thread,
            # but it shouldn't raise an error in this case
            c.close()
        self._client.close()
        self._is_closed = True

    # Server-side async methods
    def _get_async_query_status(self, token: str) -> str:
        if self.cursor_type != CursorV2:
            raise FireboltError(
                "This method is only supported for connection with service account."
            )
        cursor = self.cursor()
        cursor.execute(ASYNC_QUERY_STATUS_REQUEST.format(token=token))
        result = cursor.fetchone()
        if cursor.rowcount != 1 or not result:
            raise FireboltError("Unexpected result from async query status request.")
        columns = cursor.description
        result_dict = dict(zip([column.name for column in columns], result))
        return result_dict["status"]

    def is_async_query_running(self, token: str) -> bool:
        """
        Check if an async query is still running.

        Args:
            token: Async query token. Can be obtained from Cursor.async_query_token.

        Returns:
            bool: True if async query is still running, False otherwise
        """
        return self._get_async_query_status(token) == ASYNC_QUERY_STATUS_RUNNING

    def is_async_query_successful(self, token: str) -> Optional[bool]:
        """
        Check if an async query has finished and was successful.

        Args:
            token: Async query token. Can be obtained from Cursor.async_query_token.

        Returns:
            bool: None if the query is still running, True if successful,
                  False otherwise
        """
        status = self._get_async_query_status(token)
        if status == ASYNC_QUERY_STATUS_RUNNING:
            return None
        return status == ASYNC_QUERY_STATUS_SUCCESSFUL

    # Context manager support
    def __enter__(self) -> Connection:
        if self.closed:
            raise ConnectionClosedError("Connection is already closed.")
        return self

    def __exit__(
        self, exc_type: type, exc_val: Exception, exc_tb: TracebackType
    ) -> None:
        self.close()

    def __del__(self) -> None:
        if not self.closed:
            warn(f"Unclosed {self!r}", UserWarning)


def connect_v1(
    auth: Auth,
    user_agent_header: str,
    database: Optional[str] = None,
    account_name: Optional[str] = None,
    engine_name: Optional[str] = None,
    engine_url: Optional[str] = None,
    api_endpoint: str = DEFAULT_API_URL,
) -> Connection:
    # These parameters are optional in function signature
    # but are required to connect.
    # PEP 249 recommends making them kwargs.
    if not database:
        raise ConfigurationError("database name is required to connect.")

    validate_engine_name_and_url_v1(engine_name, engine_url)

    api_endpoint = fix_url_schema(api_endpoint)

    # Override tcp keepalive settings for connection
    no_engine_client = ClientV1(
        auth=auth,
        base_url=api_endpoint,
        account_name=account_name,
        api_endpoint=api_endpoint,
        timeout=Timeout(DEFAULT_TIMEOUT_SECONDS, read=None),
        headers={"User-Agent": user_agent_header},
    )

    # Mypy checks, this should never happen
    assert database is not None

    if not engine_name and not engine_url:
        engine_url = no_engine_client._get_database_default_engine_url(
            database=database
        )

    elif engine_name:
        engine_url = no_engine_client._resolve_engine_url(engine_name=engine_name)
    elif account_name:
        # In above if branches account name is validated since it's used to
        # resolve or get an engine url.
        # We need to manually validate account_name if none of the above
        # cases are triggered.
        assert no_engine_client.account_id is not None

    assert engine_url is not None

    engine_url = fix_url_schema(engine_url)
    client = ClientV1(
        auth=auth,
        account_name=account_name,
        api_endpoint=api_endpoint,
        timeout=Timeout(DEFAULT_TIMEOUT_SECONDS, read=None),
        headers={"User-Agent": user_agent_header},
    )
    return Connection(engine_url, database, client, CursorV1, api_endpoint)



import os
from typing import (
    Any,
    Callable,
    Dict,
    Generic,
    Optional,
    Protocol,
    Tuple,
    TypeVar,
)

T = TypeVar("T")


class ReprCacheable(Protocol):
    def __repr__(self) -> str:
        ...


def noop_if_disabled(func: Callable) -> Callable:
    """Decorator to make function do nothing if the cache is disabled."""

    def wrapper(self: "UtilCache", *args: Any, **kwargs: Any) -> Any:
        if not self.disabled:
            return func(self, *args, **kwargs)

    return wrapper


class UtilCache(Generic[T]):
    """
    Generic cache implementation to store key-value pairs.
    Created to abstract the cache implementation in case we find a better
    solution in the future.
    """

    def __init__(self, cache_name: str = "") -> None:
        self._cache: Dict[str, T] = {}
        # Allow disabling cache if we have no direct access to the constructor
        self.disabled = os.getenv("FIREBOLT_SDK_DISABLE_CACHE", False) or os.getenv(
            f"FIREBOLT_SDK_DISABLE_CACHE_${cache_name}", False
        )

    def disable(self) -> None:
        self.disabled = True

    def enable(self) -> None:
        self.disabled = False

    def get(self, key: ReprCacheable) -> Optional[T]:
        if self.disabled:
            return None
        s_key = self.create_key(key)
        return self._cache.get(s_key)

    @noop_if_disabled
    def set(self, key: ReprCacheable, value: T) -> None:
        if not self.disabled:
            s_key = self.create_key(key)
            self._cache[s_key] = value

    @noop_if_disabled
    def delete(self, key: ReprCacheable) -> None:
        s_key = self.create_key(key)
        if s_key in self._cache:
            del self._cache[s_key]

    @noop_if_disabled
    def clear(self) -> None:
        self._cache.clear()

    def create_key(self, obj: ReprCacheable) -> str:
        return repr(obj)

    def __contains__(self, key: str) -> bool:
        """Support for 'in' operator to check if key is present in cache."""
        if self.disabled:
            return False
        return key in self._cache


_firebolt_system_engine_cache = UtilCache[Tuple[str, Dict[str, str]]](
    cache_name="system_engine"
)



from __future__ import annotations

import re
from collections import namedtuple
from datetime import date, datetime, timezone
from decimal import Decimal
from enum import Enum
from io import StringIO
from typing import Any, Dict, List, Optional, Sequence, Tuple, Union

from sqlparse import parse as parse_sql  # type: ignore
from sqlparse.sql import (  # type: ignore
    Comment,
    Comparison,
    Statement,
    Token,
    TokenList,
)
from sqlparse.tokens import Comparison as ComparisonType  # type: ignore
from sqlparse.tokens import Newline  # type: ignore
from sqlparse.tokens import Whitespace  # type: ignore
from sqlparse.tokens import Token as TokenType  # type: ignore

try:
    from ciso8601 import parse_datetime  # type: ignore
except ImportError:
    unsupported_milliseconds_re = re.compile(r"(?<=\.)\d{1,5}(?!\d)")

    def _fix_milliseconds(datetime_string: str) -> str:
        # Fill milliseconds with 0 to have exactly 6 digits
        # Python parser only supports 3 or 6 digit milliseconds untill 3.11
        def align_ms(match: re.Match) -> str:
            ms = match.group()
            return ms + "0" * (6 - len(ms))

        return re.sub(unsupported_milliseconds_re, align_ms, datetime_string)

    def _fix_timezone(datetime_string: str) -> str:
        # timezone, provided as +/-dd is not supported by datetime.
        # We need to append :00 to it
        if datetime_string[-3] in "+-":
            return datetime_string + ":00"
        return datetime_string

    def parse_datetime(datetime_string: str) -> datetime:
        return datetime.fromisoformat(_fix_timezone(_fix_milliseconds(datetime_string)))


from firebolt.utils.exception import (
    DataError,
    InterfaceError,
    NotSupportedError,
)
from firebolt.utils.util import cached_property

_NoneType = type(None)
_col_types = (int, float, str, datetime, date, bool, list, Decimal, _NoneType, bytes)
# duplicating this since 3.7 can't unpack Union
ColType = Union[int, float, str, datetime, date, bool, list, Decimal, _NoneType, bytes]
RawColType = Union[int, float, str, bool, list, _NoneType]
ParameterType = Union[int, float, str, datetime, date, bool, Decimal, Sequence, bytes]

# These definitions are required by PEP-249
Date = date


def DateFromTicks(t: int) -> date:  # NOSONAR
    """Convert `ticks` to `date` for Firebolt DB."""
    return datetime.fromtimestamp(t).date()


def Time(hour: int, minute: int, second: int) -> None:  # NOSONAR
    """Unsupported: Construct `time`, for Firebolt DB."""
    raise NotSupportedError("The time construct is not supported by Firebolt")


def TimeFromTicks(t: int) -> None:  # NOSONAR
    """Unsupported: Convert `ticks` to `time` for Firebolt DB."""
    raise NotSupportedError("The time construct is not supported by Firebolt")


Timestamp = datetime
TimestampFromTicks = datetime.fromtimestamp


def Binary(value: str) -> bytes:  # NOSONAR
    """Encode a string into UTF-8."""
    return value.encode("utf-8")


STRING = str
BINARY = bytes
NUMBER = int
DATETIME = datetime
ROWID = int

Column = namedtuple(
    "Column",
    (
        "name",
        "type_code",
        "display_size",
        "internal_size",
        "precision",
        "scale",
        "null_ok",
    ),
)


class ExtendedType:
    """Base type for all extended types in Firebolt (array, decimal, struct, etc.)."""

    __name__ = "ExtendedType"

    @staticmethod
    def is_valid_type(type_: Any) -> bool:
        return type_ in _col_types or isinstance(type_, ExtendedType)

    def __hash__(self) -> int:
        return hash(str(self))


class ARRAY(ExtendedType):
    """Class for holding `array` column type information in Firebolt DB."""

    __name__ = "Array"
    _prefix = "array("

    def __init__(self, subtype: Union[type, ExtendedType]):
        if not self.is_valid_type(subtype):
            raise ValueError(f"Invalid array subtype: {str(subtype)}")
        self.subtype = subtype

    def __str__(self) -> str:
        return f"Array({str(self.subtype)})"

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, ARRAY):
            return NotImplemented
        return other.subtype == self.subtype


class DECIMAL(ExtendedType):
    """Class for holding `decimal` value information in Firebolt DB."""

    __name__ = "Decimal"
    _prefix = "Decimal("

    def __init__(self, precision: int, scale: int):
        self.precision = precision
        self.scale = scale

    def __str__(self) -> str:
        return f"Decimal({self.precision}, {self.scale})"

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, DECIMAL):
            return NotImplemented
        return other.precision == self.precision and other.scale == self.scale


class STRUCT(ExtendedType):
    __name__ = "Struct"
    _prefix = "struct("

    def __init__(self, fields: Dict[str, Union[type, ExtendedType]]):
        for name, type_ in fields.items():
            if not self.is_valid_type(type_):
                raise ValueError(f"Invalid struct field type: {str(type_)}")
        self.fields = fields

    def __str__(self) -> str:
        return f"Struct({', '.join(f'{k}: {v}' for k, v in self.fields.items())})"

    def __eq__(self, other: Any) -> bool:
        return isinstance(other, STRUCT) and other.fields == self.fields


NULLABLE_SUFFIX = "null"


class _InternalType(Enum):
    """Enum of all internal Firebolt types, except for `array`."""

    Int = "int"
    Long = "long"
    Float = "float"
    Double = "double"

    Text = "text"

    Date = "date"
    DateExt = "date_ext"
    PGDate = "pgdate"

    Timestamp = "timestamp"
    TimestampExt = "timestamp_ext"
    TimestampNtz = "timestampntz"
    TimestampTz = "timestamptz"

    Boolean = "boolean"

    Bytea = "bytea"
    Geography = "geography"

    Nothing = "Nothing"

    @cached_property
    def python_type(self) -> type:
        """Convert internal type to Python type."""
        types = {
            _InternalType.Int: int,
            _InternalType.Long: int,
            _InternalType.Float: float,
            _InternalType.Double: float,
            _InternalType.Text: str,
            _InternalType.Date: date,
            _InternalType.DateExt: date,
            _InternalType.PGDate: date,
            _InternalType.Timestamp: datetime,
            _InternalType.TimestampExt: datetime,
            _InternalType.TimestampNtz: datetime,
            _InternalType.TimestampTz: datetime,
            _InternalType.Boolean: bool,
            _InternalType.Bytea: bytes,
            _InternalType.Geography: str,
            # For simplicity, this could happen only during 'select null' query
            _InternalType.Nothing: str,
        }
        return types[self]


def split_struct_fields(raw_struct: str) -> List[str]:
    """Split raw struct inner fields string into a list of field definitions.
    >>> split_struct_fields("field1 int, field2 struct(field1 int, field2 text)")
    ['field1 int', 'field2 struct(field1 int, field2 text)']
    """
    balance = 0  # keep track of the level of nesting, and only split on level 0
    separator = ","
    res = []
    current = StringIO()
    for i, ch in enumerate(raw_struct):
        if ch == "(":
            balance += 1
        elif ch == ")":
            balance -= 1
        elif ch == separator and balance == 0:
            res.append(current.getvalue().strip())
            current = StringIO()
            continue
        current.write(ch)

    res.append(current.getvalue().strip())
    return res


def split_struct_field(raw_field: str) -> Tuple[str, str]:
    """Split raw struct field into name and type.
    >>> split_struct_field("field int")
    ('field', 'int')
    >>> split_struct_field("`with space` text null")
    ('with space', 'text null')
    >>> split_struct_field("s struct(`a b` int)")
    ('s', 'struct(`a b` int)')
    """
    raw_field = raw_field.strip()
    second_tick = (
        raw_field.find("`", raw_field.find("`") + 1)
        if raw_field.startswith("`")
        else -1
    )
    name, type_ = (
        (raw_field[: second_tick + 1], raw_field[second_tick + 1 :])
        if second_tick != -1
        else raw_field.split(" ", 1)
    )
    return name.strip(" `"), type_.strip()


def parse_type(raw_type: str) -> Union[type, ExtendedType]:  # noqa: C901
    """Parse typename provided by query metadata into Python type."""
    if not isinstance(raw_type, str):
        raise DataError(f"Invalid typename {str(raw_type)}: str expected")
    # Handle arrays
    if raw_type.startswith(ARRAY._prefix) and raw_type.endswith(")"):
        return ARRAY(parse_type(raw_type[len(ARRAY._prefix) : -1]))
    # Handle decimal
    if raw_type.startswith(DECIMAL._prefix) and raw_type.endswith(")"):
        try:
            prec_scale = raw_type[len(DECIMAL._prefix) : -1].split(",")
            precision, scale = int(prec_scale[0]), int(prec_scale[1])
            return DECIMAL(precision, scale)
        except (ValueError, IndexError):
            pass
    # Handle structs
    if raw_type.startswith(STRUCT._prefix) and raw_type.endswith(")"):
        try:
            fields_raw = split_struct_fields(raw_type[len(STRUCT._prefix) : -1])
            fields = {}
            for f in fields_raw:
                name, type_ = split_struct_field(f)
                fields[name.strip()] = parse_type(type_.strip())
            return STRUCT(fields)
        except ValueError:
            pass
    # Handle nullable
    if raw_type.endswith(NULLABLE_SUFFIX):
        return parse_type(raw_type[: -len(NULLABLE_SUFFIX)].strip(" "))
    try:
        return _InternalType(raw_type).python_type
    except ValueError:
        # Treat unknown types as strings. Better that error since user still has
        # a way to work with it
        return str


BYTEA_PREFIX = "\\x"


def _parse_bytea(str_value: str) -> bytes:
    if (
        len(str_value) < len(BYTEA_PREFIX)
        or str_value[: len(BYTEA_PREFIX)] != BYTEA_PREFIX
    ):
        raise ValueError(f"Invalid bytea value format: {BYTEA_PREFIX} prefix expected")
    return bytes.fromhex(str_value[len(BYTEA_PREFIX) :])


def parse_value(
    value: RawColType,
    ctype: Union[type, ExtendedType],
) -> ColType:
    """Provided raw value, and Python type; parses first into Python value."""
    if value is None:
        return None
    if ctype in (int, str, float):
        assert isinstance(ctype, type)  # assertion for mypy
        return ctype(value)
    if ctype is date:
        if not isinstance(value, str):
            raise DataError(f"Invalid date value {value}: str expected")
        assert isinstance(value, str)
        return parse_datetime(value).date()
    if ctype is datetime:
        if not isinstance(value, str):
            raise DataError(f"Invalid datetime value {value}: str expected")
        return parse_datetime(value)
    if ctype is bool:
        if not isinstance(value, (bool, int)):
            raise DataError(f"Invalid boolean value {value}: bool or int expected")
        return bool(value)
    if ctype is bytes:
        if not isinstance(value, str):
            raise DataError(f"Invalid bytea value {value}: str expected")
        return _parse_bytea(value)
    if isinstance(ctype, DECIMAL):
        if not isinstance(value, (str, int)):
            raise DataError(f"Invalid decimal value {value}: str or int expected")
        return Decimal(value)
    if isinstance(ctype, ARRAY):
        if not isinstance(value, list):
            raise DataError(f"Invalid array value {value}: list expected")
        return [parse_value(it, ctype.subtype) for it in value]
    if isinstance(ctype, STRUCT):
        if not isinstance(value, dict):
            raise DataError(f"Invalid struct value {value}: dict expected")
        return {
            name: parse_value(value.get(name), type_)
            for name, type_ in ctype.fields.items()
        }
    raise DataError(f"Unsupported data type returned: {ctype.__name__}")


escape_chars = {
    "\0": "\\0",
    "\\": "\\\\",
    "'": "\\'",
}


def format_value(value: ParameterType) -> str:
    """For Python value to be used in a SQL query."""
    if isinstance(value, bool):
        return "true" if value else "false"
    if isinstance(value, (int, float, Decimal)):
        return str(value)
    elif isinstance(value, str):
        return f"'{''.join(escape_chars.get(c, c) for c in value)}'"
    elif isinstance(value, datetime):
        if value.tzinfo is not None:
            value = value.astimezone(timezone.utc)
        return f"'{value.strftime('%Y-%m-%d %H:%M:%S')}'"
    elif isinstance(value, date):
        return f"'{value.isoformat()}'"
    elif isinstance(value, bytes):
        # Encode each byte into hex
        return "E'" + "".join(f"\\x{b:02x}" for b in value) + "'"
    if value is None:
        return "NULL"
    elif isinstance(value, Sequence):
        return f"[{', '.join(format_value(it) for it in value)}]"

    raise DataError(f"unsupported parameter type {type(value)}")


def format_statement(statement: Statement, parameters: Sequence[ParameterType]) -> str:
    """
    Substitute placeholders in a `sqlparse` statement with provided values.
    """
    idx = 0

    def process_token(token: Token) -> Token:
        nonlocal idx
        if token.ttype == TokenType.Name.Placeholder:
            # Replace placeholder with formatted parameter
            if idx >= len(parameters):
                raise DataError(
                    "not enough parameters provided for substitution: given "
                    f"{len(parameters)}, found one more"
                )
            formatted = format_value(parameters[idx])
            idx += 1
            return Token(TokenType.Text, formatted)
        if isinstance(token, TokenList):
            # Process all children tokens

            return TokenList([process_token(t) for t in token.tokens])
        return token

    formatted_sql = statement_to_sql(process_token(statement))

    if idx < len(parameters):
        raise DataError(
            f"too many parameters provided for substitution: given {len(parameters)}, "
            f"used only {idx}"
        )

    return formatted_sql


SetParameter = namedtuple("SetParameter", ["name", "value"])


def statement_to_set(statement: Statement) -> Optional[SetParameter]:
    """
    Try to parse `statement` as a `SET` command.
    Return `None` if it's not a `SET` command.
    """
    # Filter out meaningless tokens like Punctuation and Whitespaces
    skip_types = [Whitespace, Newline]
    tokens = [
        token
        for token in statement.tokens
        if token.ttype not in skip_types and not isinstance(token, Comment)
    ]
    # Trim tail punctuation
    right_idx = len(tokens) - 1
    while str(tokens[right_idx]) == ";":
        right_idx -= 1

    tokens = tokens[: right_idx + 1]

    # Check if it's a SET statement by checking if it starts with set
    if (
        len(tokens) > 0
        and tokens[0].ttype == TokenType.Keyword
        and tokens[0].value.lower() == "set"
    ):
        # Check if set statement has a valid format
        if len(tokens) == 2 and isinstance(tokens[1], Comparison):
            return SetParameter(
                statement_to_sql(tokens[1].left),
                statement_to_sql(tokens[1].right).strip("'"),
            )
        # Or if at least there is a comparison
        cmp_idx = next(
            (
                i
                for i, token in enumerate(tokens)
                if token.ttype == ComparisonType or isinstance(token, Comparison)
            ),
            None,
        )
        if cmp_idx:
            left_tokens, right_tokens = tokens[1:cmp_idx], tokens[cmp_idx + 1 :]
            if isinstance(tokens[cmp_idx], Comparison):
                left_tokens = left_tokens + [tokens[cmp_idx].left]
                right_tokens = [tokens[cmp_idx].right] + right_tokens

            if left_tokens and right_tokens:
                return SetParameter(
                    "".join(statement_to_sql(t) for t in left_tokens),
                    "".join(statement_to_sql(t) for t in right_tokens).strip("'"),
                )

        raise InterfaceError(
            f"Invalid set statement format: {statement_to_sql(statement)},"
            " expected SET <param> = <value>"
        )
    return None


def statement_to_sql(statement: Statement) -> str:
    return str(statement).strip().rstrip(";")


def split_format_sql(
    query: str, parameters: Sequence[Sequence[ParameterType]]
) -> List[Union[str, SetParameter]]:
    """
    Multi-statement query formatting will result in `NotSupportedError`.
    Instead, split a query into a separate statement and format with parameters.
    """
    statements = parse_sql(query)
    if not statements:
        return [query]

    if parameters:
        if len(statements) > 1:
            raise NotSupportedError(
                "Formatting multi-statement queries is not supported."
            )
        if statement_to_set(statements[0]):
            raise NotSupportedError("Formatting set statements is not supported.")
        return [format_statement(statements[0], paramset) for paramset in parameters]

    # Try parsing each statement as a SET, otherwise return as a plain sql string
    return [statement_to_set(st) or statement_to_sql(st) for st in statements]



from time import time
from typing import Optional

from firebolt.utils.exception import QueryTimeoutError


class TimeoutController:
    """A class to control the timeout of a sequence of queries.
    Used to check if the timeout has been reached and to calculate the remaining time.
    """

    timeout_error_cls = QueryTimeoutError

    def __init__(self, timeout: Optional[float]):
        self.timeout = timeout
        self.start_time = time()

    def raise_if_timeout(self) -> None:
        """Raise a TimeoutError if the timeout has been reached."""
        if self.timeout is not None and time() > (self.start_time + self.timeout):
            raise self.timeout_error_cls()

    def remaining(self) -> Optional[float]:
        """Return the remaining time before the timeout is reached."""
        return (
            None
            if self.timeout is None
            else max(0.0, self.timeout - (time() - self.start_time))
        )



import json
from typing import Any, Callable, Final, Type, TypeVar, Union

import pydantic
from typing_extensions import TypeAlias

Model = TypeVar("Model", bound="pydantic.BaseModel")

GenericCallable: TypeAlias = Callable[..., Any]

# Using `.VERSION` instead of `.__version__` for backward compatibility:
PYDANTIC_VERSION: Final[int] = int(pydantic.VERSION[0])


def use_if_version_ge(
    version_ge: int,
    obj: Union[pydantic.BaseModel, Type[Model]],
    previous_method: str,
    latest_method: str,
) -> GenericCallable:
    """
    Utility function to get desired method from base model.

    Args:
        version_ge: The version number that will be used to determine
            the desired method.
        obj: The object on which the method will be taken from
        previous_method: The method previously available in a version
            smaller than `version_ge`.
        latest_method:  The method available from `version_ge` onwards.

    """
    if PYDANTIC_VERSION >= version_ge:
        return getattr(obj, latest_method)
    else:
        return getattr(obj, previous_method)


if PYDANTIC_VERSION >= 2:
    # This import can only happen outside the BaseModel,
    # or it will raise PydanticUserError
    from pydantic import ConfigDict


class FireboltBaseModel(pydantic.BaseModel):
    if PYDANTIC_VERSION >= 2:
        # Pydantic V2 config
        model_config = ConfigDict(populate_by_name=True, from_attributes=True)

    else:
        # Using Pydantic 1.* config class for backwards compatibility
        class Config:
            extra = "forbid"
            allow_population_by_field_name = True  # Pydantic 1.8

    def model_dict(self, *args: Any, **kwargs: Any) -> dict:
        """Pydantic V2 and V1 compatible method for `dict` -> `model_dump`."""
        return use_if_version_ge(2, self, "dict", "model_dump")(*args, **kwargs)

    @classmethod
    def parse_model(cls: Type[Model], *args: Any, **kwargs: Any) -> Model:
        """Pydantic V2 and V1 compatible method for `parse_obj` -> `model_validate`."""
        return use_if_version_ge(2, cls, "parse_obj", "model_validate")(*args, **kwargs)

    def model_json(self, *args: Any, **kwargs: Any) -> str:
        """Pydantic V2 and V1 compatible method for `json` -> `model_dump_json`."""
        return use_if_version_ge(2, self, "json", "model_dump_json")(*args, **kwargs)

    def jsonable_dict(self, *args: Any, **kwargs: Any) -> dict:
        """
        Generate a dictionary representation of the service that contains serialized
        primitive types, and is therefore JSON-ready.

        This could be replaced with something native once this issue is resolved:
        https://github.com/samuelcolvin/pydantic/issues/1409

        This function is intended to improve the compatibility with HTTPX, which
        expects to take in a dictionary of primitives as input to the JSON parameter
        of its request function. See: https://www.python-httpx.org/api/#helper-functions
        """
        return json.loads(self.model_json(*args, **kwargs))



from typing import Any, List, Type

from firebolt.utils.exception import ConnectionClosedError

ASYNC_QUERY_STATUS_RUNNING = "RUNNING"
ASYNC_QUERY_STATUS_SUCCESSFUL = "ENDED_SUCCESSFULLY"
ASYNC_QUERY_STATUS_REQUEST = "CALL fb_GetAsyncStatus('{token}')"


class BaseConnection:
    def __init__(self, cursor_type: Type) -> None:
        self.cursor_type = cursor_type
        self._cursors: List[Any] = []
        self._is_closed = False

    def _remove_cursor(self, cursor: Any) -> None:
        # This way it's atomic
        try:
            self._cursors.remove(cursor)
        except ValueError:
            pass

    @property
    def closed(self) -> bool:
        """`True` if connection is closed; `False` otherwise."""
        return self._is_closed

    def commit(self) -> None:
        """Does nothing since Firebolt doesn't have transactions."""

        if self.closed:
            raise ConnectionClosedError("Unable to commit: Connection closed.")
